from math import sqrt
import numpy as np
import os

from pysph.sph.wc.transport_velocity import SummationDensity, \
    MomentumEquationPressureGradient, StateEquation,\
    MomentumEquationArtificialStress, MomentumEquationViscosity, \
    SolidWallNoSlipBC

from pysph.sph.surface_tension import InterfaceCurvatureFromNumberDensity, \
    ShadlooYildizSurfaceTensionForce, CSFSurfaceTensionForce, \
    SmoothedColor, AdamiColorGradient, MorrisColorGradient, \
    SY11DiracDelta, SY11ColorGradient, MomentumEquationViscosityAdami, \
    AdamiReproducingDivergence, CSFSurfaceTensionForceAdami,\
    MomentumEquationPressureGradientAdami, ColorGradientAdami, \
    ConstructStressMatrix, SurfaceForceAdami, SummationDensitySourceMass, \
    MomentumEquationViscosityMorris, MomentumEquationPressureGradientMorris, \
    InterfaceCurvatureFromDensity, SolidWallPressureBCnoDensity


from pysph.sph.wc.basic import TaitEOS
from pysph.sph.gas_dynamics.basic import ScaleSmoothingLength

from pysph.tools.geometry import get_2d_block, remove_overlap_particles, \
    get_2d_circle
from pysph.base.utils import get_particle_array
from pysph.base.kernels import CubicSpline, QuinticSpline
from pysph.sph.equation import Group, Equation

from pysph.solver.application import Application
from pysph.solver.solver import Solver

from pysph.sph.integrator_step import TransportVelocityStep
from pysph.sph.integrator import PECIntegrator

from pysph.base.nnps import DomainManager
from pysph.solver.utils import iter_output

dim = 2
Lx = 1.0
Ly = 1.0

nu1 = 0.05
nu2 = 0.0005
sigma = 1.0
factor1 = 0.5
factor2 = 1 / factor1
rho1 = 1.0

c0 = 20.0
gamma = 1.4
R = 287.1
rho2 = 0.001

p1 = c0**2 * rho1
p2 = c0*c0*rho2
nx = 60
dx = Lx / nx
volume = dx * dx
hdx = 1.0

h0 = hdx * dx

tf = 0.5

epsilon = 0.01 / h0
v0 = 10.0
r0 = 0.05

dt1 = 0.25*np.sqrt(rho2*h0*h0*h0/(2.0*np.pi*sigma))

dt2 = 0.25*h0/(c0+v0)

dt3 = 0.125*rho2*h0*h0/nu2

dt = 0.9*min(dt1, dt2, dt3)


d = 2


def r(x, y):
    return x*x + y*y


class MultiPhase(Application):

    def create_particles(self):
        fluid_x, fluid_y = get_2d_block(
            dx=dx, length=Lx, height=Ly, center=np.array([0., 0.]))
        rho_fluid = np.ones_like(fluid_x) * rho2
        m_fluid = rho_fluid * volume
        h_fluid = np.ones_like(fluid_x) * h0
        cs_fluid = np.ones_like(fluid_x) * c0
        circle_x, circle_y = get_2d_circle(dx=dx, r=0.2,
                                           center=np.array([0.0, 0.0]))
        rho_circle = np.ones_like(circle_x) * rho1
        m_circle = rho_circle * volume
        h_circle = np.ones_like(circle_x) * h0
        cs_circle = np.ones_like(circle_x) * c0
        wall_x, wall_y = get_2d_block(dx=dx, length=Lx+6*dx, height=Ly+6*dx,
                                      center=np.array([0., 0.]))
        rho_wall = np.ones_like(wall_x) * rho2
        m_wall = rho_wall * volume
        h_wall = np.ones_like(wall_x) * h0
        cs_wall = np.ones_like(wall_x) * c0
        additional_props = ['V', 'color', 'scolor', 'cx', 'cy', 'cz',
                            'cx2', 'cy2', 'cz2', 'nx', 'ny', 'nz', 'ddelta',
                            'uhat', 'vhat', 'what', 'auhat', 'avhat', 'awhat',
                            'ax', 'ay', 'az', 'wij', 'vmag2', 'N', 'wij_sum',
                            'rho0', 'u0', 'v0', 'w0', 'x0', 'y0', 'z0',
                            'kappa', 'arho', 'nu', 'wg', 'ug', 'vg',
                            'pi00', 'pi01', 'pi02', 'pi10', 'pi11', 'pi12',
                            'pi20', 'pi21', 'pi22']
        gas = get_particle_array(
            name='gas', x=fluid_x, y=fluid_y, h=h_fluid, m=m_fluid,
            rho=rho_fluid, cs=cs_fluid, additional_props=additional_props)
        gas.nu[:] = nu2
        gas.color[:] = 0.0
        liquid = get_particle_array(
            name='liquid', x=circle_x, y=circle_y, h=h_circle, m=m_circle,
            rho=rho_circle, cs=cs_circle, additional_props=additional_props)
        liquid.nu[:] = nu1
        liquid.color[:] = 1.0
        wall = get_particle_array(
            name='wall', x=wall_x, y=wall_y, h=h_wall, m=m_wall,
            rho=rho_wall, cs=cs_wall, additional_props=additional_props)
        wall.color[:] = 0.0
        remove_overlap_particles(wall, liquid, dx_solid=dx, dim=2)
        remove_overlap_particles(wall, gas, dx_solid=dx, dim=2)
        remove_overlap_particles(gas, liquid, dx_solid=dx, dim=2)
        gas.add_output_arrays(['V', 'color', 'cx', 'cy', 'nx', 'ny', 'ddelta',
                               'kappa', 'N', 'scolor', 'p'])
        liquid.add_output_arrays(['V', 'color', 'cx', 'cy', 'nx', 'ny',
                                  'ddelta', 'kappa', 'N', 'scolor', 'p'])
        wall.add_output_arrays(['V', 'color', 'cx', 'cy', 'nx', 'ny',
                                'ddelta', 'kappa', 'N', 'scolor', 'p'])
        for i in range(len(gas.x)):
            R = sqrt(r(gas.x[i], gas.y[i]) + 0.0001*gas.h[i]*gas.h[i])
            f = np.exp(-R/r0)
            gas.u[i] = v0*gas.x[i]*(1.0-(gas.y[i]*gas.y[i])/(r0*R))*f/r0
            gas.v[i] = -v0*gas.y[i]*(1.0-(gas.x[i]*gas.x[i])/(r0*R))*f/r0
        for i in range(len(liquid.x)):
            R = sqrt(r(liquid.x[i], liquid.y[i]) +
                     0.0001*liquid.h[i]*liquid.h[i])
            liquid.u[i] = v0*liquid.x[i] * \
                (1.0 - (liquid.y[i]*liquid.y[i])/(r0*R))*f/r0
            liquid.v[i] = -v0*liquid.y[i] * \
                (1.0 - (liquid.x[i]*liquid.x[i])/(r0*R))*f/r0
        return [liquid, gas, wall]

    def create_solver(self):
        kernel = QuinticSpline(dim=2)
        integrator = PECIntegrator(liquid=TransportVelocityStep(),
                                   gas=TransportVelocityStep())
        solver = Solver(
            kernel=kernel, dim=dim, integrator=integrator,
            dt=dt, tf=tf, adaptive_timestep=False)
        return solver

    def create_equations(self):
        adami_stress_equations = [
            Group(equations=[
                SummationDensity(
                    dest='liquid', sources=[
                        'liquid', 'wall', 'gas']),
                SummationDensity(
                    dest='gas', sources=[
                        'liquid', 'wall', 'gas']),
                SummationDensity(
                    dest='wall', sources=['liquid', 'wall', 'gas'])
            ]),
            Group(equations=[
                TaitEOS(dest='liquid', sources=None, rho0=rho1, c0=c0, gamma=1,
                        p0=p1),
                TaitEOS(dest='gas', sources=None,
                        rho0=rho2, c0=c0, gamma=1, p0=p1),
                SolidWallPressureBCnoDensity(dest='wall', sources=['liquid',
                                                                   'gas']),
            ]),
            Group(equations=[
                ColorGradientAdami(dest='liquid', sources=['liquid', 'wall',
                                                           'gas']),
                ColorGradientAdami(dest='gas', sources=[
                              'liquid', 'wall', 'gas']),
            ]),
            Group(equations=[ConstructStressMatrix(dest='liquid', sources=None,
                                                   sigma=sigma, d=2),
                             ConstructStressMatrix(dest='gas', sources=None,
                                                   sigma=sigma, d=2)]),
            Group(
                equations=[
                    MomentumEquationPressureGradientAdami(
                        dest='liquid', sources=['liquid', 'wall', 'gas']),
                    MomentumEquationPressureGradientAdami(
                        dest='gas', sources=['liquid', 'wall', 'gas']),
                    MomentumEquationViscosityAdami(
                        dest='liquid', sources=['liquid', 'gas']),
                    MomentumEquationViscosityAdami(
                        dest='gas', sources=['liquid', 'gas']),
                    SurfaceForceAdami(
                        dest='liquid', sources=['liquid', 'wall', 'gas']),
                    SurfaceForceAdami(
                        dest='gas', sources=['liquid', 'wall', 'gas']),
                    SolidWallNoSlipBC(dest='liquid', sources=['wall'], nu=nu1),
                    SolidWallNoSlipBC(dest='gas', sources=['wall'], nu=nu2),
                ]),
                ]
        return adami_stress_equations

    def post_process(self):
        try:
            import matplotlib
            matplotlib.use('Agg')
            import matplotlib.pyplot as plt
        except ImportError:
            print("Post processing requires Matplotlib")
            return
        from pysph.solver.utils import load
        files = self.output_files
        t = []
        centerx = []
        centery = []
        for f in files:
            data = load(f)
            pa = data['arrays']['liquid']
            t.append(data['solver_data']['t'])
            x = pa.x
            y = pa.y
            length = len(x)
            cx = 0
            cy = 0
            count = 0
            for i in range(length):
                if x[i] > 0 and y[i] > 0:
                    cx += x[i]
                    cy += y[i]
                    count += 1
                else:
                    continue
            # As the masses are all the same in this case
            centerx.append(cx/count)
            centery.append(cy/count)
        fname = os.path.join(self.output_dir, 'results.npz')
        np.savez(fname, t=t, centerx=centerx, centery=centery)
        plt.plot(t, centerx, label='x position')
        plt.plot(t, centery, label='y position')
        plt.legend()
        fig1 = os.path.join(self.output_dir, 'centerofmassposvst')
        plt.savefig(fig1)
        plt.close()


if __name__ == '__main__':
    app = MultiPhase()
    app.run()
    app.post_process()



"""Tests for the PySPH parallel module"""

import shutil
import tempfile
import unittest

import numpy as np
from pytest import mark, importorskip
from pysph.tools import run_parallel_script

path = run_parallel_script.get_directory(__file__)


class ParticleArrayTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        importorskip("pysph.parallel.parallel_manager")

    def test_get_strided_indices(self):
        # Given
        from pysph.parallel.parallel_manager import get_strided_indices

        indices = np.array([1, 5, 3])

        # When
        idx = get_strided_indices(indices, 1)
        # Then
        np.testing.assert_array_equal(idx, indices)

        # When
        idx = get_strided_indices(indices, 2)
        # Then
        np.testing.assert_array_equal(
            idx, [2, 3, 10, 11, 6, 7]
        )

        # When
        idx = get_strided_indices(indices, 3)
        # Then
        np.testing.assert_array_equal(
            idx, [3, 4, 5, 15, 16, 17, 9, 10, 11]
        )


class ParticleArrayExchangeTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        importorskip("mpi4py.MPI")
        importorskip("pyzoltan.core.zoltan")

    @mark.parallel
    def test_lb_exchange(self):
        run_parallel_script.run(filename='lb_exchange.py', nprocs=4, path=path)

    @mark.parallel
    def test_remote_exchange(self):
        run_parallel_script.run(
            filename='remote_exchange.py', nprocs=4, path=path
        )


class SummationDensityTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        importorskip("mpi4py.MPI")
        importorskip("pyzoltan.core.zoltan")

    @mark.slow
    @mark.parallel
    def test_summation_density(self):
        run_parallel_script.run(
            filename='summation_density.py', nprocs=4, path=path
        )


class MPIReduceArrayTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        importorskip("mpi4py.MPI")
        importorskip("pyzoltan.core.zoltan")

    def setUp(self):
        self.root = tempfile.mkdtemp()

    def tearDown(self):
        shutil.rmtree(self.root)

    @mark.parallel
    def test_mpi_reduce_array(self):
        run_parallel_script.run(
            filename='reduce_array.py', nprocs=4, path=path
        )

    @mark.parallel
    def test_parallel_reduce(self):
        args = ['--directory=%s' % self.root]
        run_parallel_script.run(
            filename='simple_reduction.py', args=args, nprocs=4, path=path,
            timeout=60.0
        )


class DumpLoadTestCase(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        importorskip("mpi4py.MPI")
        importorskip("pyzoltan.core.zoltan")

    @mark.parallel
    def test_dump_and_load_work_in_parallel(self):
        run_parallel_script.run(
            filename='check_dump_load.py', nprocs=4, path=path
        )


if __name__ == '__main__':
    unittest.main()



""" Helper functions to generate commonly used geometries.

PySPH used an axis convention as follows:

Y
|
|
|
|
|
|      /Z
|     /
|    /
|   /
|  /
| /
|/_________________X



"""

import numpy
from numpy import concatenate, where, array
from pysph.base.utils import get_particle_array_wcsph, get_particle_array_iisph
from cyarray.api import LongArray


def create_2D_tank(x1, y1, x2, y2, dx):
    """ Generate an open rectangular tank.

    Parameters:
    -----------

    x1,y1,x2,y2 : Coordinates defining the rectangle in 2D

    dx : The spacing to use

    """

    yl = numpy.arange(y1, y2 + dx / 2, dx)
    xl = numpy.ones_like(yl) * x1
    nl = len(xl)

    yr = numpy.arange(y1, y2 + dx / 2, dx)
    xr = numpy.ones_like(yr) * x2
    nr = len(xr)

    xb = numpy.arange(x1 + dx, x2 - dx + dx / 2, dx)
    yb = numpy.ones_like(xb) * y1
    nb = len(xb)

    x = numpy.concatenate([xl, xb, xr])
    y = numpy.concatenate([yl, yb, yr])

    return x, y


def create_2D_filled_region(x1, y1, x2, y2, dx):
    x, y = numpy.mgrid[x1:x2 + dx / 2:dx, y1:y2 + dx / 2:dx]
    x = x.ravel()
    y = y.ravel()

    return x, y


def create_obstacle(x1, x2, height, dx):
    eps = 1e-6

    # left inside wall
    yli = numpy.arange(dx / 2.0, height + eps, dx)
    xli = numpy.ones_like(yli) * x1

    # left outside wall
    ylo = numpy.arange(dx, height + dx / 2.0 + eps, dx)
    xlo = numpy.ones_like(ylo) * x1 - dx / 2.0

    # # right inside wal
    # yri = numpy.arange( dx/2.0, height + eps, dx )
    # xri = numpy.ones_like(yri) * x2

    # # right outter wall
    # yro = numpy.arange( dx, height + dx/2.0 + eps, dx )
    # xro = numpy.ones_like(yro) * x2 + dx/2.0

    # concatenate the arrays
    x = numpy.concatenate((xli, xlo))
    y = numpy.concatenate((yli, ylo))

    return x, y


class DamBreak2DGeometry(object):
    def __init__(self, container_width=4.0, container_height=3.0,
                 fluid_column_width=1.0, fluid_column_height=2.0,
                 dx=0.03, dy=0.03, nboundary_layers=4,
                 ro=1000.0, co=1.0, with_obstacle=False,
                 beta=1.0, nfluid_offset=2, hdx=1.5, iisph=False,
                 wall_hex_pack=True):

        self.container_width = container_width
        self.container_height = container_height
        self.fluid_column_height = fluid_column_height
        self.fluid_column_width = fluid_column_width

        self.nboundary_layers = nboundary_layers
        self.nfluid_offset = nfluid_offset
        self.beta = beta
        self.hdx = hdx
        self.dx = dx
        self.dy = dy
        self.iisph = iisph
        self.wall_hex_pack = wall_hex_pack

        self.nsolid = 0
        self.nfluid = 0

        self.ro = ro
        self.co = co

        self.with_obstacle = with_obstacle

    def get_wall(self, nboundary_layers=4):
        container_width = self.container_width
        container_height = self.container_height

        dx, dy = self.dx / self.beta, self.dy / self.beta
        factor = 0.5 if self.wall_hex_pack else 1.0
        _x = []
        _y = []
        for i in range(nboundary_layers):
            xb, yb = create_2D_tank(
                x1=-factor * i * dx, y1=-factor * i * dy,
                x2=container_width + factor * i * dx, y2=container_height, dx=dx)

            _x.append(xb)
            _y.append(yb)

        x = numpy.concatenate(_x)
        y = numpy.concatenate(_y)
        self.nsolid = len(x)

        return x, y

    def get_fluid(self, noffset=2):
        fluid_column_width = self.fluid_column_width
        fluid_column_height = self.fluid_column_height

        dx, dy = self.dx, self.dy
        factor = 0.5

        _x = []
        _y = []
        for i in range(noffset):
            ii = i + 1
            xf, yf = create_2D_filled_region(
                x1=dx - factor * i * dx, y1=dx - factor * i * dx,
                # x1=0.5*ii*dx, y1=0.5*ii*dx,
                x2=fluid_column_width + factor * i * dx, y2=fluid_column_height,
                dx=dx)

            _x.append(xf)
            _y.append(yf)

        x = numpy.concatenate(_x)
        y = numpy.concatenate(_y)
        self.nfluid = len(x)

        return x, y

    def create_particles(self, nboundary_layers=2, nfluid_offset=2,
                         hdx=1.5, **kwargs):
        nfluid = self.nfluid
        xf, yf = self.get_fluid(nfluid_offset)
        if self.iisph:
            fluid = get_particle_array_iisph(name='fluid', x=xf, y=yf)
        else:
            fluid = get_particle_array_wcsph(name='fluid', x=xf, y=yf)

        fluid.gid[:] = list(range(fluid.get_number_of_particles()))

        np = nfluid

        xb, yb = self.get_wall(nboundary_layers)
        if self.iisph:
            boundary = get_particle_array_iisph(name='boundary', x=xb, y=yb)
        else:
            boundary = get_particle_array_wcsph(name='boundary', x=xb, y=yb)

        np += boundary.get_number_of_particles()

        dx, dy, ro = self.dx, self.dy, self.ro

        # smoothing length, mass and density
        fluid.h[:] = numpy.ones_like(xf) * hdx * dx
        if nfluid_offset == 2:
            fluid.m[:] = dx * dy * ro * 0.5
        else:
            fluid.m[:] = dx * dy * ro
        fluid.rho[:] = ro
        if not self.iisph:
            fluid.rho0[:] = ro

        boundary.h[:] = numpy.ones_like(xb) * hdx * dx
        if nboundary_layers == 2:
            boundary.m[:] = dx * dy * ro * 0.5
        else:
            boundary.m[:] = dx * dy * ro
        boundary.rho[:] = ro
        if not self.iisph:
            boundary.rho0[:] = ro

        # create the particles list
        particles = [fluid, boundary]

        if self.with_obstacle:
            xo, yo = create_obstacle(x1=2.5, x2=2.5 + dx, height=0.25, dx=dx)
            gido = numpy.array(list(range(xo.size)), dtype=numpy.uint32)

            obstacle = get_particle_array_wcsph(name='obstacle', x=xo, y=yo)

            obstacle.h[:] = numpy.ones_like(xo) * hdx * dx
            obstacle.m[:] = dx * dy * 0.5 * ro
            obstacle.rho[:] = ro
            if not self.iisph:
                obstacle.rho0[:] = ro

            # add the obstacle to the boundary particles
            boundary.append_parray(obstacle)

            np += obstacle.get_number_of_particles()

        # set the gid for the boundary particles
        boundary.gid[:] = list(range(boundary.get_number_of_particles()))

        # boundary particles can do with a reduced list of properties
        # to be saved to disk since they are fixed
        boundary.set_output_arrays(
            ['x', 'y', 'rho', 'm', 'h', 'p', 'tag', 'pid', 'gid'])
        if self.iisph:
            boundary.add_output_arrays(['V'])

        print("2D dam break with %d fluid, %d boundary particles" % (
            fluid.get_number_of_particles(),
            boundary.get_number_of_particles()))

        return particles


class DamBreak3DGeometry(object):
    def __init__(
            self, container_height=1.0, container_width=1.0, container_length=3.22,
            fluid_column_height=0.55, fluid_column_width=1.0, fluid_column_length=1.228,
            obstacle_center_x=2.5, obstacle_center_y=0,
            obstacle_length=0.16, obstacle_height=0.161, obstacle_width=0.4,
            nboundary_layers=5, with_obstacle=True, dx=0.02, hdx=1.2, rho0=1000.0):

        # save the geometry details
        self.container_width = container_width
        self.container_length = container_length
        self.container_height = container_height

        self.fluid_column_length = fluid_column_length
        self.fluid_column_width = fluid_column_width
        self.fluid_column_height = fluid_column_height

        self.obstacle_center_x = obstacle_center_x
        self.obstacle_center_y = obstacle_center_y

        self.obstacle_width = obstacle_width
        self.obstacle_length = obstacle_length
        self.obstacle_height = obstacle_height

        self.nboundary_layers = nboundary_layers
        self.dx = dx

        self.hdx = hdx
        self.rho0 = rho0
        self.with_obstacle = with_obstacle

    def get_max_speed(self, g=9.81):
        return numpy.sqrt(2 * g * self.fluid_column_height)

    def create_particles(self, **kwargs):
        fluid_column_height = self.fluid_column_height
        fluid_column_width = self.fluid_column_width
        fluid_column_length = self.fluid_column_length

        container_height = self.container_height
        container_length = self.container_length
        container_width = self.container_width

        obstacle_height = self.obstacle_height
        obstacle_length = self.obstacle_length
        obstacle_width = self.obstacle_width

        obstacle_center_x = self.obstacle_center_x
        obstacle_center_y = self.obstacle_center_y

        nboundary_layers = self.nboundary_layers
        dx = self.dx

        # get the domain limits
        ghostlims = nboundary_layers * dx

        xmin, xmax = 0.0 - ghostlims, container_length + ghostlims
        zmin, zmax = 0.0 - ghostlims, container_height + ghostlims

        cw2 = 0.5 * container_width
        ymin, ymax = -cw2 - ghostlims, cw2 + ghostlims

        # create all particles
        eps = 0.1 * dx
        xx, yy, zz = numpy.mgrid[xmin:xmax + eps:dx,
                                 ymin:ymax + eps:dx,
                                 zmin:zmax + eps:dx]

        x = xx.ravel()
        y = yy.ravel()
        z = zz.ravel()

        # create a dummy particle array from which we'll sort
        pa = get_particle_array_wcsph(name='block', x=x, y=y, z=z)

        # get the individual arrays
        indices = []
        findices = []
        oindices = []

        obw2 = 0.5 * obstacle_width
        obl2 = 0.5 * obstacle_length
        obh = obstacle_height
        ocx = obstacle_center_x
        ocy = obstacle_center_y

        for i in range(x.size):
            xi = x[i]
            yi = y[i]
            zi = z[i]

            # fluid
            if ((0 < xi <= fluid_column_length) and
                (-cw2 < yi < cw2) and
                    (0 < zi <= fluid_column_height)):

                findices.append(i)

            # obstacle
            if ((ocx - obl2 <= xi <= ocx + obl2) and
                (ocy - obw2 <= yi <= ocy + obw2) and
                    (0 < zi <= obh)):

                oindices.append(i)

        # extract the individual arrays
        fa = LongArray(len(findices))
        fa.set_data(numpy.array(findices))
        fluid = pa.extract_particles(fa)
        fluid.set_name('fluid')

        if self.with_obstacle:
            oa = LongArray(len(oindices))
            oa.set_data(numpy.array(oindices))
            obstacle = pa.extract_particles(oa)
            obstacle.set_name('obstacle')

        indices = concatenate((where(y <= -cw2)[0],
                               where(y >= cw2)[0],
                               where(x >= container_length)[0],
                               where(x <= 0)[0],
                               where(z <= 0)[0]))

        # remove duplicates
        indices = array(list(set(indices)))

        wa = LongArray(indices.size)
        wa.set_data(indices)
        boundary = pa.extract_particles(wa)
        boundary.set_name('boundary')

        # create the particles
        if self.with_obstacle:
            particles = [fluid, boundary, obstacle]
        else:
            particles = [fluid, boundary]

        # set up particle properties
        h0 = self.hdx * dx

        volume = dx**3
        m0 = self.rho0 * volume

        for pa in particles:
            pa.m[:] = m0
            pa.h[:] = h0

            pa.rho[:] = self.rho0

        nf = fluid.get_number_of_particles()
        nb = boundary.get_number_of_particles()

        if self.with_obstacle:
            no = obstacle.get_number_of_particles()
            print(
                "3D dam break with %d fluid, %d boundary, "
                "%d obstacle particles" %
                (nf, nb, no))
        else:
            print(
                "3D dam break with %d fluid, %d boundary particles" %
                (nf, nb))

        # load balancing props for the arrays
        # fluid.set_lb_props(['x', 'y', 'z', 'u', 'v', 'w', 'rho', 'h', 'm',
        #                   'gid', 'x0', 'y0', 'z0', 'u0', 'v0', 'w0', 'rho0'])
        fluid.set_lb_props(list(fluid.properties.keys()))

        #boundary.set_lb_props(['x', 'y', 'z', 'rho', 'h', 'm', 'gid', 'rho0'])
        #obstacle.set_lb_props(['x', 'y', 'z', 'rho', 'h', 'm', 'gid', 'rho0'])
        boundary.set_lb_props(list(boundary.properties.keys()))

        # boundary and obstacle particles can do with a reduced list of
        # properties to be saved to disk since they are fixed
        boundary.set_output_arrays(
            ['x', 'y', 'z', 'rho', 'm', 'h', 'p', 'tag', 'pid', 'gid'])

        if self.with_obstacle:
            obstacle.set_lb_props(list(obstacle.properties.keys()))
            obstacle.set_output_arrays(
                ['x', 'y', 'z', 'rho', 'm', 'h', 'p', 'tag', 'pid', 'gid'])

        return particles



"""do-nothing outlet first used in SPH by

    - Federico, Ivan, et al. "Simulating 2D open-channel flows through an SPH
      model." European Journal of Mechanics-B/Fluids 34 (2012): 35-46.

"""

from pysph.sph.integrator import PECIntegrator
from pysph.sph.bc.inlet_outlet_manager import InletOutletManager
from pysph.sph.wc.edac import EDACScheme


class SimpleInletOutlet(InletOutletManager):
    def add_io_properties(self, pa, scheme=None):
        DEFAULT_PROPS = [
            'disp', 'ioid', 'Bp', 'A', 'wij', 'po', 'uho', 'vho', 'who',
            'Buh', 'Bvh', 'Bwh', 'x0', 'y0', 'z0', 'uhat', 'vhat', 'what',
            'xn', 'yn', 'zn']
        STRIDE_DATA = {
            'A': 16, 'Bu': 4, 'Bv': 4, 'Bw': 4, 'Bp': 4, 'Brho': 4, 'uo':
            4, 'vo': 4, 'wo': 4, 'po': 4, 'rhoo': 4, 'Bau': 4, 'Bav': 4,
            'Baw': 4, 'auo': 4, 'avo': 4, 'awo': 4, 'Buh': 4, 'Bvh': 4,
            'Bwh': 4, 'uho': 4, 'vho': 4, 'who': 4}
        for prop in DEFAULT_PROPS:
            if prop in STRIDE_DATA:
                pa.add_property(prop, stride=STRIDE_DATA[prop])
            else:
                pa.add_property(prop)

        pa.add_constant('uref', 0.0)

    def get_stepper(self, scheme, cls, edactvf=False):

        from pysph.sph.bc.inlet_outlet_manager import (
            InletStep, OutletStepWithUhat)
        steppers = {}
        if (cls == PECIntegrator):
            if isinstance(scheme, EDACScheme):
                for inlet in self.inlets:
                    steppers[inlet] = InletStep()
                for outlet in self.outlets:
                    steppers[outlet] = OutletStepWithUhat()
                for g_inlet in self.ghost_inlets:
                    steppers[g_inlet] = InletStep()
                self.active_stages = [2]

        return steppers

    def get_equations(self, scheme=None, summation_density=False,
                      edactvf=False):
        from pysph.sph.equation import Group
        from pysph.sph.bc.interpolate import (
            UpdateMomentMatrix, EvaluateUhat, EvaluateP, ExtrapolateUhat,
            ExtrapolateP, CopyUhatFromGhost, CopyPFromGhost)
        from pysph.sph.bc.inlet_outlet_manager import (
            UpdateNormalsAndDisplacements, CopyNormalsandDistances)

        equations = []
        g00 = []
        i = 0
        for info in self.inletinfo:
            g00.append(UpdateNormalsAndDisplacements(
                dest=info.pa_name, sources=None, xn=info.normal[0],
                yn=info.normal[1], zn=info.normal[2], xo=info.refpoint[0],
                yo=info.refpoint[1], zo=info.refpoint[2]
                ))
            g00.append(CopyNormalsandDistances(
                dest=self.inlet_pairs[info.pa_name], sources=[info.pa_name]))
            i = i + 1

        equations.append(Group(equations=g00, real=False))
        g02 = []
        for name in self.ghost_inlets:
            g02.append(UpdateMomentMatrix(
                dest=name, sources=self.fluids, dim=self.dim))

        equations.append(Group(equations=g02, real=False))

        g03 = []
        for name in self.ghost_inlets:
            g03.append(EvaluateUhat(dest=name, sources=self.fluids,
                                    dim=self.dim))
            g03.append(EvaluateP(dest=name, sources=self.fluids,
                                 dim=self.dim))

        equations.append(Group(equations=g03, real=False))

        g04 = []
        for name in self.ghost_inlets:
            g04.append(ExtrapolateUhat(dest=name, sources=None))
            g04.append(ExtrapolateP(dest=name, sources=None))

        equations.append(Group(equations=g04, real=False))

        g05 = []
        for io in self.inlet_pairs.keys():
            g05.append(CopyUhatFromGhost(
                    dest=io, sources=[self.inlet_pairs[io]]))
            g05.append(CopyPFromGhost(
                dest=io, sources=[self.inlet_pairs[io]]))

        equations.append(Group(equations=g05, real=False))

        g06 = []
        for inlet in self.inletinfo:
            for eqn in inlet.equations:
                g06.append(eqn)
        for outlet in self.outletinfo:
            for eqn in outlet.equations:
                g06.append(eqn)

        equations.append(Group(equations=g06, real=False))

        return equations



"""
Incompressible SPH.
The divergence free formulation of ISPH is implemented here.
See "An SPH Projection Method", Cummins S.J., Rudman M, JCOMP(1999).

Taylor-Green example only works with --shift-freq=1.
"""
import numpy
from compyle.api import declare
from pysph.sph.scheme import Scheme, add_bool_argument
from pysph.base.utils import get_particle_array
from pysph.sph.integrator import Integrator
from pysph.sph.integrator_step import IntegratorStep
from pysph.sph.equation import Equation, Group, MultiStageEquations


def get_particle_array_isph(constants=None, **props):
    isph_props = [
        'u0', 'v0', 'w0', 'x0', 'y0', 'z0', 'rho0', 'diag', 'rhs',
        'V0', 'V', 'au', 'av', 'aw'
    ]

    # No of particles
    N = len(props['gid'])
    consts = {
        'np': numpy.array([N], dtype=int),
    }

    if constants:
        consts.update(constants)

    pa = get_particle_array(
        additional_props=isph_props, constants=consts, **props
    )
    pa.add_property('ctr', type='int')
    pa.add_property('coeff', stride=100)
    pa.add_property('col_idx', stride=100, type='int')
    pa.add_property('row_idx', stride=100, type='int')
    pa.add_output_arrays(['p'])
    return pa


class ISPHIntegrator(Integrator):
    def one_timestep(self, t, dt):
        self.initialize()

        self.compute_accelerations(0)

        self.stage1()

        self.update_domain()

        self.do_post_stage(0.5*dt, 1)

        self.compute_accelerations(1)

        self.stage2()

        self.update_domain()

        self.do_post_stage(dt, 2)

    def initial_acceleration(self, t, dt):
        pass


class ISPHStep(IntegratorStep):
    def initialize(self, d_idx, d_x, d_y, d_z, d_x0, d_y0, d_z0, d_u, d_v,
                   d_w, d_u0, d_v0, d_w0, dt, d_rho0, d_rho, d_V):
        d_x0[d_idx] = d_x[d_idx]
        d_y0[d_idx] = d_y[d_idx]
        d_z0[d_idx] = d_z[d_idx]

        d_u0[d_idx] = d_u[d_idx]
        d_v0[d_idx] = d_v[d_idx]
        d_w0[d_idx] = d_w[d_idx]

        d_rho0[d_idx] = d_rho[d_idx]

    def stage1(self, d_idx, d_x, d_y, d_z, d_u, d_v, d_w, d_au, d_av, d_aw,
               d_V0, d_V, dt):
        d_x[d_idx] += dt*d_u[d_idx]
        d_y[d_idx] += dt*d_v[d_idx]
        d_z[d_idx] += dt*d_w[d_idx]

        d_u[d_idx] += dt*d_au[d_idx]
        d_v[d_idx] += dt*d_av[d_idx]
        d_w[d_idx] += dt*d_aw[d_idx]

        d_V0[d_idx] = d_V[d_idx]

    def stage2(self, d_idx, d_x, d_y, d_z, d_u, d_v, d_w, d_u0, d_v0, d_w0,
               d_x0, d_y0, d_z0, dt, d_au, d_av, d_aw):
        d_u[d_idx] += dt*d_au[d_idx]
        d_v[d_idx] += dt*d_av[d_idx]
        d_w[d_idx] += dt*d_aw[d_idx]

        d_x[d_idx] = d_x0[d_idx] + 0.5*dt * (d_u[d_idx] + d_u0[d_idx])
        d_y[d_idx] = d_y0[d_idx] + 0.5*dt * (d_v[d_idx] + d_v0[d_idx])
        d_z[d_idx] = d_z0[d_idx] + 0.5*dt * (d_w[d_idx] + d_w0[d_idx])


class MomentumEquationBodyForce(Equation):
    def __init__(self, dest, sources, gx=0.0, gy=0.0, gz=0.0):
        self.gx = gx
        self.gy = gy
        self.gz = gz
        super(MomentumEquationBodyForce, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def post_loop(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] += self.gx
        d_av[d_idx] += self.gy
        d_aw[d_idx] += self.gz


class VelocityDivergence(Equation):
    def initialize(self, d_idx, d_rhs):
        d_rhs[d_idx] = 0.0

    def loop(self, d_idx, s_idx, s_m, s_rho, d_rhs, dt, VIJ, DWIJ):
        Vj = s_m[s_idx] / s_rho[s_idx]
        vdotdwij = VIJ[0]*DWIJ[0] + VIJ[1]*DWIJ[1] + VIJ[2]*DWIJ[2]
        d_rhs[d_idx] += -Vj * vdotdwij / dt


class VelocityDivergenceDFDI(Equation):
    def initialize(self, d_idx, d_rhs):
        d_rhs[d_idx] = 0.0

    def loop(self, d_idx, s_idx, s_m, s_rho, d_rhs, dt, VIJ, DWIJ):
        Vj = s_m[s_idx] / s_rho[s_idx]
        vdotdwij = VIJ[0]*DWIJ[0] + VIJ[1]*DWIJ[1] + VIJ[2]*DWIJ[2]
        d_rhs[d_idx] += -2*Vj * vdotdwij / dt


class DensityInvariance(Equation):
    def __init__(self, dest, sources, rho0):
        self.rho0 = rho0
        super(DensityInvariance, self).__init__(dest, sources)

    def post_loop(self, d_idx, d_rho, d_rhs, dt):
        rho0 = self.rho0
        d_rhs[d_idx] = (rho0 - d_rho[d_idx]) / (dt*dt*rho0)


class DensityInvarianceDFDI(Equation):
    def post_loop(self, d_idx, d_V, d_V0, d_rhs, dt):
        V0 = d_V0[d_idx]
        d_rhs[d_idx] = 2*(V0 - d_V[d_idx]) / (dt*dt*V0)


class PressureCoeffMatrix(Equation):
    def initialize(self, d_idx, d_ctr, d_diag, d_col_idx):
        # Make only the diagonals zero as the rest are not summed.
        d_diag[d_idx] = 0.0
        d_ctr[d_idx] = 0

        # Initialize col_idx to -1 so as to use it in cond while constructing
        # pressure coeff matrix.
        i = declare('int')
        for i in range(100):
            d_col_idx[d_idx*100 + i] = -1

    def loop(self, d_idx, s_idx, s_m, d_rho, s_rho, d_gid, d_coeff, d_ctr,
             d_col_idx, d_row_idx, d_diag, XIJ, DWIJ, R2IJ, EPS):
        rhoij = (s_rho[s_idx] + d_rho[d_idx])
        rhoij2_1 = 1.0/(rhoij*rhoij)

        xdotdwij = XIJ[0]*DWIJ[0] + XIJ[1]*DWIJ[1] + XIJ[2]*DWIJ[2]

        fac = 8.0 * s_m[s_idx] * rhoij2_1 * xdotdwij / (R2IJ + EPS)

        j, k = declare('int', 3)
        j = d_gid[s_idx]

        d_diag[d_idx] += fac

        k = int(d_ctr[d_idx])
        d_coeff[d_idx*100 + k] = -fac
        d_col_idx[d_idx*100 + k] = j
        d_row_idx[d_idx*100 + k] = d_idx
        d_ctr[d_idx] += 1


class PPESolve(Equation):
    def py_initialize(self, dst, t, dt):
        import scipy.sparse as sp
        from scipy.sparse.linalg import bicgstab

        coeff = declare('object')
        cond = declare('object')
        n = declare('int')
        n = dst.np[0]

        # Mask all indices which are not used in the construction.
        cond = (dst.col_idx != -1)

        coeff = sp.csr_matrix(
            (dst.coeff[cond], (dst.row_idx[cond], dst.col_idx[cond])),
            shape=(n, n)
        )
        # Add tiny random noise so the matrix is not singular.
        cond = abs(dst.rhs) > 1e-9
        dst.diag[cond] -= numpy.random.random(n)[cond]

        coeff += sp.diags(dst.diag)

        # Pseudo-Neumann boundary conditions
        dst.rhs[cond] -= dst.rhs[cond].mean()

        dst.p[:], ec = bicgstab(coeff, dst.rhs, x0=dst.p)
        assert ec == 0, "Not converging!"


class MomentumEquationPressureGradient(Equation):
    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, s_idx, s_m, d_p, s_p, d_rho, s_rho, d_au, d_av, d_aw,
             DWIJ):
        Vj = s_m[s_idx] / s_rho[s_idx]
        pij = (d_p[d_idx] - s_p[s_idx])
        fac = Vj * pij / d_rho[d_idx]

        d_au[d_idx] += fac * DWIJ[0]
        d_av[d_idx] += fac * DWIJ[1]
        d_aw[d_idx] += fac * DWIJ[2]


class MomentumEquationPressureGradientSymmetric(Equation):
    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, s_idx, s_m, d_p, s_p, d_rho, s_rho, d_au, d_av, d_aw,
             DWIJ):
        rhoi2 = d_rho[d_idx]*d_rho[d_idx]
        rhoj2 = s_rho[s_idx]*s_rho[s_idx]
        pij = d_p[d_idx]/rhoi2 + s_p[s_idx]/rhoj2
        fac = -s_m[s_idx] * pij

        d_au[d_idx] += fac * DWIJ[0]
        d_av[d_idx] += fac * DWIJ[1]
        d_aw[d_idx] += fac * DWIJ[2]


class UpdatePosition(Equation):
    def post_loop(self, d_idx, d_au, d_av, d_aw, d_x, d_y, d_z, dt):
        d_x[d_idx] += d_au[d_idx] * dt*dt * 0.5
        d_y[d_idx] += d_av[d_idx] * dt*dt * 0.5
        d_z[d_idx] += d_aw[d_idx] * dt*dt * 0.5


class CheckDensityError(Equation):
    def __init__(self, dest, sources, rho0, tol=0.01):
        self.conv = 0
        self.rho0 = rho0
        self.tol = tol
        self.count = 0
        self.rho_err = 0
        super(CheckDensityError, self).__init__(dest, sources)

    def py_initialize(self, dst, t, dt):
        self.rho_err = numpy.abs(dst.rho - self.rho0).max()
        self.conv = 1 if self.rho_err < self.tol else -1
        self.count += 1

    def converged(self):
        return self.conv


class FreeSurfaceBoundaryCondition(Equation):
    def initialize(self, d_rho, d_rho0, d_rhs, d_diag, d_idx, d_coeff, d_ctr,
                   d_col_idx, d_row_idx):
        i = declare('int')
        if d_rho[d_idx]/d_rho0[d_idx] < 0.98:
            d_rhs[d_idx] = 0.0
            d_diag[d_idx] = 1.0
            d_ctr[d_idx] = 1
            for i in range(100):
                d_coeff[d_idx*100 + i] = 0.0
                d_col_idx[d_idx*100 + i] = -1
                d_row_idx[d_idx*100 + i] = d_idx


class MomentumEquationPressureGradientSymmetricMirror(Equation):
    def initialize(self, d_idx, d_au, d_av, d_aw):
        d_au[d_idx] = 0.0
        d_av[d_idx] = 0.0
        d_aw[d_idx] = 0.0

    def loop(self, d_idx, s_idx, s_m, d_p, s_p, d_rho, s_rho, d_au, d_av, d_aw,
             s_rho0, DWIJ, XIJ, RIJ, HIJ, SPH_KERNEL):
        rhoi2 = d_rho[d_idx]*d_rho[d_idx]
        rhoj2 = s_rho[s_idx]*s_rho[s_idx]
        pij = d_p[d_idx]/rhoi2 + s_p[s_idx]/rhoj2
        fac = -s_m[s_idx] * pij
        xij = declare('matrix(3)')
        dwij = declare('matrix(3)')

        if s_rho[s_idx]/s_rho0[s_idx] < 0.98:
            rhoi2 = d_rho[d_idx]*d_rho[d_idx]
            rhoj2 = s_rho[s_idx]*s_rho[s_idx]
            # Shao and Lao mirror condition, Pj = -Pi.
            pij = d_p[d_idx]/rhoi2 - d_p[d_idx]/rhoj2
            fac = -s_m[s_idx] * pij

            xij[0] = XIJ[0]*2
            xij[1] = XIJ[1]*2
            xij[2] = XIJ[2]*2

            SPH_KERNEL.gradient(xij, 2*RIJ, HIJ, dwij)

            d_au[d_idx] += fac * dwij[0]
            d_av[d_idx] += fac * dwij[1]
            d_aw[d_idx] += fac * dwij[2]
        else:
            d_au[d_idx] += fac * DWIJ[0]
            d_av[d_idx] += fac * DWIJ[1]
            d_aw[d_idx] += fac * DWIJ[2]


class ISPHScheme(Scheme):
    def __init__(self, fluids, solids, dim, nu, rho0, c0, alpha, beta=0.0,
                 gx=0.0, gy=0.0, gz=0.0, tolerance=0.01, symmetric=False):
        self.fluids = fluids
        self.solver = None
        self.dim = dim
        self.nu = nu
        self.gx = gx
        self.gy = gy
        self.gz = gz
        self.c0 = c0
        self.alpha = alpha
        self.beta = beta
        self.tolerance = tolerance
        self.rho0 = rho0
        self.symmetric = symmetric

    def add_user_options(self, group):
        group.add_argument(
            '--alpha', action='store', type=float, dest='alpha',
            default=None,
            help='Artificial viscosity.'
        )
        add_bool_argument(
            group, 'symmetric', dest='symmetric', default=None,
            help='Use symmetric form of pressure gradient.'
        )

    def consume_user_options(self, options):
        _vars = ['alpha', 'symmetric']
        data = dict((var, self._smart_getattr(options, var))
                    for var in _vars)
        self.configure(**data)

    def configure_solver(self, kernel=None, integrator_cls=None,
                         extra_steppers=None, **kw):
        import pysph.base.kernels as kern
        if kernel is None:
            kernel = kern.QuinticSpline(dim=self.dim)
        steppers = {}
        if extra_steppers is not None:
            steppers.update(extra_steppers)

        step_cls = ISPHStep
        for fluid in self.fluids:
            if fluid not in steppers:
                steppers[fluid] = step_cls()

        if integrator_cls is not None:
            cls = integrator_cls
        else:
            cls = ISPHIntegrator

        integrator = cls(**steppers)

        from pysph.solver.solver import Solver
        self.solver = Solver(
            dim=self.dim, integrator=integrator, kernel=kernel, **kw
        )

    def _get_viscous_eqns(self):
        from pysph.sph.wc.transport_velocity import (
            MomentumEquationArtificialViscosity
        )
        from pysph.sph.wc.viscosity import LaminarViscosity

        all = self.fluids
        eq, stg = [], []
        for fluid in self.fluids:
            eq.append(
                LaminarViscosity(dest=fluid, sources=self.fluids,
                                 nu=self.nu)
            )
            eq.append(
                MomentumEquationArtificialViscosity(
                    dest=fluid, sources=self.fluids, c0=self.c0,
                    alpha=self.alpha
                )
            )
            eq.append(
                MomentumEquationBodyForce(
                    dest=fluid, sources=self.fluids, gx=self.gx, gy=self.gy,
                    gz=self.gz)
            )
        stg.append(Group(equations=eq))
        return stg

    def _get_ppe(self):

        all = self.fluids
        eq2, stg = [], []
        for fluid in self.fluids:
            eq2.append(VelocityDivergence(dest=fluid, sources=all))
            eq2.append(PressureCoeffMatrix(dest=fluid, sources=all))
        stg.append(Group(equations=eq2))

        eq22 = []
        for fluid in self.fluids:
            eq22.append(PPESolve(dest=fluid, sources=all))
        stg.append(Group(equations=eq22))
        return stg

    def get_equations(self):
        all = self.fluids

        all_eqns = []
        # Compute Viscous and Body forces
        stg1 = self._get_viscous_eqns()
        all_eqns.append(stg1)

        stg2 = self._get_ppe()

        # Compute acceleration due to pressure, initialize au/av/aw to 0.
        eq4 = []
        for fluid in self.fluids:
            if self.symmetric:
                eq4.append(
                    MomentumEquationPressureGradientSymmetric(
                        dest=fluid, sources=all
                    )
                )
            else:
                eq4.append(
                    MomentumEquationPressureGradient(dest=fluid, sources=all)
                )
        stg2.append(Group(equations=eq4))
        all_eqns.append(stg2)

        return MultiStageEquations(all_eqns)

    def setup_properties(self, particles, clean=True):
        particle_arrays = dict([(p.name, p) for p in particles])
        dummy = get_particle_array_isph(name='junk',
                                        gid=particle_arrays['fluid'].gid)
        props = []
        for x, arr in dummy.properties.items():
            tmp = dict(name=x, type=arr.get_c_type())
            if x in dummy.stride:
                tmp.update(stride=dummy.stride[x])
            props.append(tmp)
        constants = [dict(name=x, data=v) for x, v in dummy.constants.items()]
        output_props = dummy.output_property_arrays
        for fluid in self.fluids:
            pa = particle_arrays[fluid]
            self._ensure_properties(pa, props, clean)
            pa.set_output_arrays(output_props)
            for const in constants:
                pa.add_constant(**const)



import numpy as np
from numpy import cos, sin, cosh, sinh

# SPH equations
from pysph.sph.solid_mech.basic import (ElasticSolidsScheme,
                                        get_particle_array_elastic_dynamics)

from pysph.base.kernels import CubicSpline
from pysph.solver.application import Application


class OscillatingPlate(Application):
    def initialize(self):
        self.L = 0.2
        self.H = 0.02
        # wave number K
        self.KL = 1.875
        self.K = 1.875 / self.L

        # edge velocity of the plate (m/s)
        self.Vf = 0.05
        self.dx_plate = 0.002
        self.h = 1.3 * self.dx_plate
        self.plate_rho0 = 1000.
        self.plate_E = 2. * 1e6
        self.plate_nu = 0.3975

        self.plate_inside_wall_length = self.L / 4.
        self.wall_layers = 3

        self.alpha = 1.0
        self.beta = 1.0
        self.xsph_eps = 0.5
        self.artificial_stress_eps = 0.3

        self.tf = 1.0
        self.dt = 1e-5

    def create_plate(self):
        dx = self.dx_plate
        xp, yp = np.mgrid[-self.plate_inside_wall_length:self.L +
                          dx / 2.:dx, -self.H / 2.:self.H / 2. + dx / 2.:dx]
        xp = xp.ravel()
        yp = yp.ravel()
        return xp, yp

    def create_wall(self):
        xp, yp = self.create_plate()

        # get the minimum and maximum of the plate
        xp_min = xp.min()
        yp_min = yp.min()
        yp_max = yp.max()
        dx = self.dx_plate
        xw_upper, yw_upper = np.mgrid[
            -self.plate_inside_wall_length:self.dx_plate / 2.:dx, yp_max +
            dx:yp_max + dx + (self.wall_layers - 1) * dx + dx / 2.:dx]
        xw_upper = xw_upper.ravel()
        yw_upper = yw_upper.ravel()

        xw_lower, yw_lower = np.mgrid[
            -self.plate_inside_wall_length:self.dx_plate / 2.:dx, yp_min -
            dx:yp_min - dx - (self.wall_layers - 1) * dx - dx / 2.:-dx]
        xw_lower = xw_lower.ravel()
        yw_lower = yw_lower.ravel()

        xw_left_max = xp_min - dx
        xw_left_min = xw_left_max - (self.wall_layers - 1) * dx - dx / 2.
        yw_left_max = yw_upper.max() + dx / 2.
        yw_left_min = yw_lower.min()

        xw_left, yw_left = np.mgrid[xw_left_max:xw_left_min:-dx, yw_left_min:
                                    yw_left_max:dx]
        xw_left = xw_left.ravel()
        yw_left = yw_left.ravel()

        # wall coordinates
        xw, yw = np.concatenate((xw_lower, xw_upper, xw_left)), np.concatenate(
            (yw_lower, yw_upper, yw_left))

        return xw, yw

    def create_particles(self):
        xp, yp = self.create_plate()
        m = self.plate_rho0 * self.dx_plate**2.

        # get the index of the particle which will be used to compute the
        # amplitude
        xp_max = max(xp)
        fltr = np.argwhere(xp == xp_max)
        fltr_idx = int(len(fltr) / 2.)
        amplitude_idx = fltr[fltr_idx][0]

        kernel = CubicSpline(dim=2)
        self.wdeltap = kernel.kernel(rij=self.dx_plate, h=self.h)
        plate = get_particle_array_elastic_dynamics(
            x=xp, y=yp, m=m, h=self.h,
            rho=self.plate_rho0, name="plate", constants=dict(
                wdeltap=self.wdeltap, n=4, rho_ref=self.plate_rho0,
                E=self.plate_E, nu=self.plate_nu, amplitude_idx=amplitude_idx))

        ##################################
        # vertical velocity of the plate #
        ##################################
        # initialize with zero at the beginning
        v = np.zeros_like(xp)
        v = v.ravel()

        # set the vertical velocity for particles which are only
        # out of the wall
        K = self.K
        # L = self.L
        KL = self.KL
        M = sin(KL) + sinh(KL)
        N = cos(KL) + cosh(KL)
        Q = 2 * (cos(KL) * sinh(KL) - sin(KL) * cosh(KL))

        # import pudb; pudb.set_trace()
        fltr = xp > 0
        tmp1 = (cos(K * xp[fltr]) - cosh(K * xp[fltr]))
        tmp2 = (sin(K * xp[fltr]) - sinh(K * xp[fltr]))
        v[fltr] = self.Vf * plate.cs[0] * (M * tmp1 - N * tmp2) / Q

        # set vertical velocity
        plate.v = v

        # #########################################
        # #### Create the wall particle array #####
        # #########################################
        # create the particle array
        xw, yw = self.create_wall()
        wall = get_particle_array_elastic_dynamics(
            x=xw, y=yw, m=m, h=self.h, rho=self.plate_rho0, name="wall",
            constants=dict(E=self.plate_E, nu=self.plate_nu))

        return [plate, wall]

    def create_scheme(self):
        s = ElasticSolidsScheme(elastic_solids=['plate'], solids=['wall'],
                                dim=2)
        s.configure_solver(dt=self.dt, tf=self.tf)
        return s

    def post_process(self):
        if len(self.output_files) == 0:
            return

        from pysph.solver.utils import iter_output

        files = self.output_files
        t, amplitude = [], []
        for sd, array in iter_output(files, 'plate'):
            _t = sd['t']
            t.append(_t)
            amplitude.append(array.y[array.amplitude_idx[0]])

        import matplotlib
        import os
        matplotlib.use('Agg')

        from matplotlib import pyplot as plt
        plt.clf()
        plt.plot(t, amplitude)
        plt.xlabel('t')
        plt.ylabel('Amplitude')
        plt.legend()
        fig = os.path.join(self.output_dir, "amplitude.png")
        plt.savefig(fig, dpi=300)


if __name__ == '__main__':
    app = OscillatingPlate()
    app.run()
    app.post_process()



"""Woodward Collela blastwave (2 minutes)

Two discontinuities moving towards each other and the
results after they interact
"""
from pysph.sph.scheme import (
    GSPHScheme, SchemeChooser, ADKEScheme, GasDScheme
)
from pysph.sph.gas_dynamics.psph import PSPHScheme
from pysph.sph.gas_dynamics.tsph import TSPHScheme
from pysph.sph.wc.crksph import CRKSPHScheme
from pysph.base.utils import get_particle_array as gpa
from pysph.base.nnps import DomainManager
from pysph.solver.application import Application
import numpy

# Numerical constants
dim = 1
gamma = 1.4
gamma1 = gamma - 1.0

# solution parameters
dt = 5e-6
tf = 0.038


class WCBlastwave(Application):
    def initialize(self):
        self.xmin = 0.0
        self.xmax = 1.0
        self.domain_length = self.xmax - self.xmin
        self.rho = 1.0
        self.p1 = 1000
        self.p2 = 0.01
        self.p3 = 100
        self.u = 0.0
        self.gamma = gamma
        self.hdx = 1.5
        self.n_particles = 1000

    def add_user_options(self, group):
        group.add_argument(
            "--hdx", action="store", type=float,
            dest="hdx", default=1.5,
            help="Ratio h/dx."
        )
        group.add_argument("--nparticles", action="store", type=float,
                           dest="nprt", default=1000,
                           help="Number of particles")

    def consume_user_options(self):
        self.n_particles = self.options.nprt

    def create_particles(self):
        self.dx = self.domain_length / self.n_particles
        x = numpy.arange(
            self.xmin + self.dx*0.5, self.xmax, self.dx
            )

        p = numpy.ones_like(x) * self.p2

        left_indices = numpy.where(x < 0.1)[0]
        right_indices = numpy.where(x > 0.9)[0]

        p[left_indices] = self.p1
        p[right_indices] = self.p3

        h = self.hdx*self.dx
        m = self.dx * self.rho
        e = p / ((self.gamma - 1) * self.rho)

        cs = numpy.sqrt(self.gamma * p / self.rho)

        fluid = gpa(
            name='fluid', x=x, rho=self.rho, p=p, h=h, m=m, e=e, cs=cs,
            h0=h, u=0
            )

        self.scheme.setup_properties([fluid])
        return [fluid]

    def create_domain(self):
        return DomainManager(
            xmin=self.xmin, xmax=self.xmax,
            mirror_in_x=True
        )

    def create_scheme(self):
        self.dt = dt
        self.tf = tf
        adke = ADKEScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            alpha=1, beta=1.0, k=1.0, eps=0.8, g1=0.2, g2=0.4,
            has_ghosts=True)

        mpm = GasDScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            kernel_factor=1.2, alpha1=1.0, alpha2=0.1,
            beta=2.0, update_alpha1=True, update_alpha2=True,
            has_ghosts=True
        )

        gsph = GSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            kernel_factor=1.0,
            g1=0.2, g2=0.4, rsolver=2, interpolation=1, monotonicity=1,
            interface_zero=True, hybrid=False, blend_alpha=2.0,
            niter=20, tol=1e-6, has_ghosts=True
        )

        crk = CRKSPHScheme(
            fluids=['fluid'], dim=dim, rho0=0, c0=0, nu=0, h0=0, p0=0,
            gamma=gamma, cl=4, cq=1, eta_crit=0.2, has_ghosts=True
        )

        # Use 400 particles
        psph = PSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            hfact=1.2
        )

        # Use 400 particles
        tsph = TSPHScheme(
            fluids=['fluid'], solids=[], dim=dim, gamma=gamma,
            hfact=1.2
        )

        s = SchemeChooser(
            default='gsph', gsph=gsph, adke=adke, mpm=mpm, crksph=crk,
            psph=psph, tsph=tsph
            )
        return s

    def configure_scheme(self):
        s = self.scheme
        if self.options.scheme == 'mpm':
            s.configure(kernel_factor=1.2)
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=True, pfreq=50)
        elif self.options.scheme == 'adke':
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)
        elif self.options.scheme == 'gsph':
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)
        elif self.options.scheme == 'crksph':
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=20)
        elif self.options.scheme in ['tsph', 'psph']:
            s.configure(hfact=1.2)
            s.configure_solver(dt=self.dt, tf=self.tf,
                               adaptive_timestep=False, pfreq=50)

    def post_process(self):
        if len(self.output_files) < 1 or self.rank > 0:
            return
        try:
            import matplotlib
            matplotlib.use('Agg')
            from matplotlib import pyplot
        except ImportError:
            print("Post processing requires matplotlib.")
            return
        from pysph.solver.utils import load
        import os

        plot_exact = False
        plot_legends = ['pysph(%s)' % self.options.scheme]

        try:
            import h5py
            plot_exact = True
            plot_legends.append('exact')
        except ImportError:
            print("h5py not found, exact data will not be plotted")

        fname = os.path.join(
            os.path.dirname(__file__), 'wc_exact.hdf5'
        )
        props = ["rho", "u", "p", "e"]
        props_h5 = ["'" + pr + "'" for pr in props]
        if plot_exact:
            h5file = h5py.File(fname)
            dataset = h5file['data_0']

        outfile = self.output_files[-1]
        data = load(outfile)
        pa = data['arrays']['fluid']
        x = pa.x
        u = pa.u
        e = pa.e
        p = pa.p
        rho = pa.rho
        prop_vals = [rho, u, p, e]
        for _i, prop in enumerate(props):
            pyplot.plot(x, prop_vals[_i])
            if plot_exact:
                pyplot.scatter(
                    dataset.get(props_h5[_i])['data_0'].get("'x'")
                    ['data_0'][:],
                    dataset.get(props_h5[_i])['data_0'].get("'data'")
                    ['data_0'][:],
                    c='k', s=4
                    )
            pyplot.xlabel('x')
            pyplot.ylabel(props[_i])
            pyplot.legend(plot_legends)
            fig = os.path.join(self.output_dir, props[_i] + ".png")
            pyplot.savefig(fig, dpi=300)
            pyplot.close('all')

        fname = os.path.join(self.output_dir, 'results.npz')
        numpy.savez(fname, x=pa.x, u=pa.u, e=pa.e, cs=pa.cs, rho=pa.rho,
                    p=pa.p, h=pa.h)


if __name__ == '__main__':
    app = WCBlastwave()
    app.run()
    app.post_process()



from pysph.base.tree.tree import Tree
from pysph.base.tree.helpers import ParticleArrayWrapper, get_helper, \
    make_vec_dict, ctype_to_dtype, get_vector_dtype
from compyle.opencl import profile_kernel, DeviceWGSException, get_queue, \
    named_profile, get_context
from compyle.array import Array
from pytools import memoize

import sys
import numpy as np

import pyopencl as cl
from pyopencl.scan import GenericScanKernel
import pyopencl.tools

from mako.template import Template

# For Mako
disable_unicode = False if sys.version_info.major > 2 else True


class IncompatibleTreesException(Exception):
    pass


@named_profile('neighbor_count_prefix_sum', backend='opencl')
@memoize
def _get_neighbor_count_prefix_sum_kernel(ctx):
    return GenericScanKernel(ctx, np.int32,
                             arguments="__global int *ary",
                             input_expr="ary[i]",
                             scan_expr="a+b", neutral="0",
                             output_statement="ary[i] = prev_item")


@memoize
def _get_macros_preamble(c_type, sorted, dim):
    result = Template("""
    #define IN_BOUNDS(X, MIN, MAX) ((X >= MIN) && (X < MAX))
    #define NORM2(X, Y, Z) ((X)*(X) + (Y)*(Y) + (Z)*(Z))
    #define NORM2_2D(X, Y) ((X)*(X) + (Y)*(Y))
    #define MIN(X, Y) ((X) < (Y) ? (X) : (Y))
    #define MAX(X, Y) ((X) > (Y) ? (X) : (Y))
    #define AVG(X, Y) (((X) + (Y)) / 2)
    #define ABS(X) ((X) > 0 ? (X) : -(X))
    #define SQR(X) ((X) * (X))

    typedef struct {
        union {
            float s0;
            float x;
        };
    } float1;

    typedef struct {
        union {
            double s0;
            double x;
        };
    } double1;

    % if sorted:
    #define PID(idx) (idx)
    % else:
    #define PID(idx) (pids[idx])
    % endif

    char contains(${data_t}${dim} node_xmin1, ${data_t}${dim} node_xmax1,
                  ${data_t}${dim} node_xmin2, ${data_t}${dim} node_xmax2)
    {
        // Check if node n1 contains node n2
        char res = 1;
        % for i in range(dim):
            res = res && (node_xmin1.s${i} <= node_xmin2.s${i}) &&
                  (node_xmax1.s${i} >= node_xmax2.s${i});
        % endfor

        return res;
    }

    char contains_search(${data_t}${dim} node_xmin1,
                  ${data_t}${dim} node_xmax1,
                  ${data_t} node_hmax1,
                  ${data_t}${dim} node_xmin2, ${data_t}${dim} node_xmax2)
    {
        // Check if node n1 contains node n2 with n1 having
        // its search radius extension
        ${data_t} h = node_hmax1;
        char res = 1;
        %for i in range(dim):
            res = res & (node_xmin1.s${i} - h <= node_xmin2.s${i}) &
                  (node_xmax1.s${i} + h >= node_xmax2.s${i});
        %endfor

        return res;
    }

    char intersects(${data_t}${dim} node_xmin1, ${data_t}${dim} node_xmax1,
                    ${data_t} node_hmax1,
                    ${data_t}${dim} node_xmin2, ${data_t}${dim} node_xmax2,
                    ${data_t} node_hmax2) {
        // Check if node n1 'intersects' node n2
        ${data_t} cdist;
        ${data_t} w1, w2, wavg = 0;
        char res = 1;
        ${data_t} h = MAX(node_hmax1, node_hmax2);

        % for i in range(dim):
            cdist = fabs((node_xmin1.s${i} + node_xmax1.s${i}) / 2 -
                         (node_xmin2.s${i} + node_xmax2.s${i}) / 2);
            w1 = fabs(node_xmin1.s${i} - node_xmax1.s${i});
            w2 = fabs(node_xmin2.s${i} - node_xmax2.s${i});
            wavg = AVG(w1, w2);
            res &= (cdist - wavg <= h);
        % endfor

        return res;
    }
    """, disable_unicode=disable_unicode).render(data_t=c_type, sorted=sorted,
                                                 dim=dim)
    return result


@memoize
def _get_node_bound_kernel_parameters(dim, data_t, xvars):
    result = {}
    result['setup'] = Template(
        r"""
        ${data_t} xmin[${dim}] = {${', '.join(['INFINITY'] * dim)}};
        ${data_t} xmax[${dim}] = {${', '.join(['-INFINITY'] * dim)}};
        ${data_t} hmax = 0;
        """, disable_unicode=disable_unicode).render(dim=dim, data_t=data_t)

    result['args'] = Template(
        r"""int *pids,
        % for v in xvars:
        ${data_t} *${v},
        % endfor
        ${data_t} *h,
        ${data_t} radius_scale,
        ${data_t}${dim} *node_xmin,
        ${data_t}${dim} *node_xmax,
        ${data_t} *node_hmax
        """, disable_unicode=disable_unicode).render(dim=dim,
                                                     data_t=data_t,
                                                     xvars=xvars)

    result['leaf_operation'] = Template(
        r"""
        for (int j=pbound.s0; j < pbound.s1; j++) {
        int pid = PID(j);
        % for d in range(dim):
            xmin[${d}] = fmin(xmin[${d}], ${xvars[d]}[pid]);
            xmax[${d}] = fmax(xmax[${d}], ${xvars[d]}[pid]);
        % endfor
        hmax = fmax(h[pid] * radius_scale, hmax);
        }
        """, disable_unicode=disable_unicode).render(dim=dim, xvars=xvars)

    result['node_operation'] = Template(
        r"""
        % for i in range(2 ** dim):
            % for d in range(dim):
                xmin[${d}] = fmin(
                    xmin[${d}], node_xmin[child_offset + ${i}].s${d}
                );
                xmax[${d}] = fmax(
                    xmax[${d}], node_xmax[child_offset + ${i}].s${d}
                );
            % endfor
            hmax = fmax(hmax, node_hmax[child_offset + ${i}]);
        % endfor
        """, disable_unicode=disable_unicode).render(dim=dim)

    result['output_expr'] = Template(
        """
        % for d in range(dim):
            node_xmin[node_idx].s${d} = xmin[${d}];
            node_xmax[node_idx].s${d} = xmax[${d}];
        % endfor
        node_hmax[node_idx] = hmax;
        """, disable_unicode=disable_unicode).render(dim=dim)

    return result


@memoize
def _get_leaf_neighbor_kernel_parameters(data_t, dim, args, setup, operation,
                                         output_expr):
    result = {
        'setup': Template(r"""
            ${data_t}${dim} node_xmin1;
            ${data_t}${dim} node_xmax1;
            ${data_t} node_hmax1;

            ${data_t}${dim} node_xmin2;
            ${data_t}${dim} node_xmax2;
            ${data_t} node_hmax2;

            node_xmin1 = node_xmin_dst[cid_dst];
            node_xmax1 = node_xmax_dst[cid_dst];
            node_hmax1 = node_hmax_dst[cid_dst];

            %(setup)s;
            """ % dict(setup=setup),
                          disable_unicode=disable_unicode).render(
            data_t=data_t, dim=dim),
        'node_operation': Template("""
            node_xmin2 = node_xmin_src[cid_src];
            node_xmax2 = node_xmax_src[cid_src];
            node_hmax2 = node_hmax_src[cid_src];

            if (!intersects(node_xmin1, node_xmax1, node_hmax1,
                            node_xmin2, node_xmax2, node_hmax2) &&
                !contains(node_xmin2, node_xmax2, node_xmin1, node_xmax1)) {
                flag = 0;
                break;
            }
            """, disable_unicode=disable_unicode).render(data_t=data_t),
        'leaf_operation': Template("""
            node_xmin2 = node_xmin_src[cid_src];
            node_xmax2 = node_xmax_src[cid_src];
            node_hmax2 = node_hmax_src[cid_src];

            if (intersects(node_xmin1, node_xmax1, node_hmax1,
                            node_xmin2, node_xmax2, node_hmax2) ||
                contains_search(node_xmin1, node_xmax1, node_hmax1,
                                node_xmin2, node_xmax2)) {
                %(operation)s;
            }
            """ % dict(operation=operation),
                                   disable_unicode=disable_unicode).render(),
        'output_expr': output_expr,
        'args': Template("""
            ${data_t}${dim} *node_xmin_src, ${data_t}${dim} *node_xmax_src,
            ${data_t} *node_hmax_src,
            ${data_t}${dim} *node_xmin_dst, ${data_t}${dim} *node_xmax_dst,
            ${data_t} *node_hmax_dst,
            """ + args, disable_unicode=disable_unicode).render(data_t=data_t,
                                                                dim=dim)

    }
    return result


# Support for 1D
def register_custom_pyopencl_ctypes():
    cl.tools.get_or_register_dtype('float1', np.dtype([('s0', np.float32)]))
    cl.tools.get_or_register_dtype('double1', np.dtype([('s0', np.float64)]))


register_custom_pyopencl_ctypes()


class PointTree(Tree):
    def __init__(self, pa, dim=2, leaf_size=32, radius_scale=2.0,
                 use_double=False, c_type='float'):
        super(PointTree, self).__init__(pa.get_number_of_particles(), 2 ** dim,
                                        leaf_size)

        assert (1 <= dim <= 3)
        self.max_depth = None
        self.dim = dim
        self.powdim = 2 ** self.dim
        self.xvars = ('x', 'y', 'z')[:dim]

        self.c_type = c_type
        self.c_type_src = 'double' if use_double else 'float'

        if use_double and c_type == 'float':
            # Extend the search radius a little to account for rounding errors
            radius_scale = radius_scale * (1 + 2e-7)

        # y and z coordinates need to be present for 1D and z for 2D
        # This is because the NNPS implementation below assumes them to be
        # just set to 0.
        self.pa = ParticleArrayWrapper(pa, self.c_type_src,
                                       self.c_type, ('x', 'y', 'z', 'h'))

        self.radius_scale = radius_scale
        self.use_double = use_double

        self.helper = get_helper('tree/point_tree.mako', self.c_type)
        self.xmin = None
        self.xmax = None
        self.hmin = None
        self.make_vec = make_vec_dict[c_type][self.dim]
        self.ctx = get_context()

    def set_index_function_info(self):
        self.index_function_args = ["sfc"]
        self.index_function_arg_ctypes = ["ulong"]
        self.index_function_arg_dtypes = [np.uint64]
        self.index_function_consts = ['mask', 'rshift']
        self.index_function_const_ctypes = ['ulong', 'char']
        self.index_code = "((sfc[i] & mask) >> rshift)"

    def _calc_cell_size_and_depth(self):
        self.cell_size = self.hmin * self.radius_scale * (1. + 1e-3)

        # Logic from gpu_domain_manager.py
        if self.cell_size < 1e-6:
            self.cell_size = 1

        # This lets the tree grow up to log2(128) = 7 layers beyond what it
        # could have previously. Pretty arbitrary.
        self.cell_size /= 128
        max_width = max((self.xmax[i] - self.xmin[i]) for i in range(self.dim))
        self.max_depth = int(np.ceil(np.log2(max_width / self.cell_size))) + 1

    def _bin(self):
        dtype = ctype_to_dtype(self.c_type)
        fill_particle_data = self.helper.get_kernel("fill_particle_data",
                                                    dim=self.dim,
                                                    xvars=self.xvars)
        pa_gpu = self.pa.gpu
        args = [getattr(pa_gpu, v).dev for v in self.xvars]
        args += [dtype(self.cell_size),
                 self.make_vec(*[self.xmin[i] for i in range(self.dim)]),
                 self.sfc.dev, self.pids.dev]
        fill_particle_data(*args)

    def get_index_constants(self, depth):
        rshift = np.uint8(self.dim * (self.max_depth - depth - 1))
        mask = np.uint64((2 ** self.dim - 1) << rshift)
        return mask, rshift

    def _adjust_domain_width(self):
        # Convert width of domain to a power of 2 multiple of cell size
        # (Optimal width for cells)
        # Note that this makes sure the width in _all_ dimensions is the
        # same. We want our nodes to be cubes ideally.
        cell_size = self.hmin * self.radius_scale * (1. + 1e-5)
        max_width = np.max(self.xmax - self.xmin)

        new_width = cell_size * 2.0 ** int(
            np.ceil(np.log2(max_width / cell_size)))

        diff = (new_width - (self.xmax - self.xmin)) / 2

        self.xmin -= diff
        self.xmax += diff

    def setup_build(self, xmin, xmax, hmin):
        self._setup_build()
        self.pa.sync()
        self.xmin = np.array(xmin)
        self.xmax = np.array(xmax)
        self.hmin = hmin
        self._adjust_domain_width()
        self._calc_cell_size_and_depth()
        self._bin()

    def build(self, fixed_depth=None):
        self._build(self.max_depth if fixed_depth is None else fixed_depth)
        self._get_unique_cids_and_count()

    def refresh(self, xmin, xmax, hmin, fixed_depth=None):
        self.setup_build(xmin, xmax, hmin)
        self.build(fixed_depth)

    def _sort(self):
        """Set tree as being sorted

        The particle array needs to be aligned by the caller!
        """
        if not self.sorted:
            self.pa.sync()
            self.sorted = 1

    ###########################################################################
    # General algorithms
    ###########################################################################
    def set_node_bounds(self):
        vector_data_t = get_vector_dtype(self.c_type, self.dim)
        dtype = ctype_to_dtype(self.c_type)

        self.node_xmin = self.allocate_node_prop(vector_data_t)
        self.node_xmax = self.allocate_node_prop(vector_data_t)
        self.node_hmax = self.allocate_node_prop(dtype)

        params = _get_node_bound_kernel_parameters(self.dim, self.c_type,
                                                   self.xvars)
        set_node_bounds = self.tree_bottom_up(
            params['args'], params['setup'], params['leaf_operation'],
            params['node_operation'], params['output_expr'],
            preamble=_get_macros_preamble(self.c_type, self.sorted, self.dim)
        )
        set_node_bounds = profile_kernel(set_node_bounds, 'set_node_bounds',
                                         backend='opencl')

        pa_gpu = self.pa.gpu
        dtype = ctype_to_dtype(self.c_type)

        args = [self, self.pids.dev]
        args += [getattr(pa_gpu, v).dev for v in self.xvars]
        args += [pa_gpu.h.dev,
                 dtype(self.radius_scale),
                 self.node_xmin.dev, self.node_xmax.dev,
                 self.node_hmax.dev]

        set_node_bounds(*args)

    ###########################################################################
    # Nearest Neighbor Particle Search (NNPS)
    ###########################################################################
    def _leaf_neighbor_operation(self, tree_src, args, setup, operation,
                                 output_expr):
        # Template for finding neighboring cids of a cell.
        params = _get_leaf_neighbor_kernel_parameters(self.c_type, self.dim,
                                                      args,
                                                      setup, operation,
                                                      output_expr)
        kernel = tree_src.leaf_tree_traverse(
            params['args'], params['setup'], params['node_operation'],
            params['leaf_operation'], params['output_expr'],
            preamble=_get_macros_preamble(self.c_type, self.sorted, self.dim)
        )

        def callable(*args):
            return kernel(tree_src, self,
                          tree_src.node_xmin.dev,
                          tree_src.node_xmax.dev,
                          tree_src.node_hmax.dev,
                          self.node_xmin.dev, self.node_xmax.dev,
                          self.node_hmax.dev,
                          *args)

        return callable

    def find_neighbor_cids(self, tree_src):
        neighbor_cid_count = Array(np.uint32, n=self.unique_cid_count + 1,
                                   backend='opencl')
        find_neighbor_cid_counts = self._leaf_neighbor_operation(
            tree_src,
            args="uint2 *pbounds, int *cnt",
            setup="int count=0",
            operation="""
                    if (pbounds[cid_src].s0 < pbounds[cid_src].s1)
                        count++;
                    """,
            output_expr="cnt[i] = count;"
        )
        find_neighbor_cid_counts = profile_kernel(
            find_neighbor_cid_counts, 'find_neighbor_cid_count',
            backend='opencl'
        )
        find_neighbor_cid_counts(tree_src.pbounds.dev,
                                 neighbor_cid_count.dev)

        neighbor_psum = _get_neighbor_count_prefix_sum_kernel(self.ctx)
        neighbor_psum(neighbor_cid_count.dev)

        total_neighbors = int(neighbor_cid_count.dev[-1].get())
        neighbor_cids = Array(np.uint32, n=total_neighbors,
                              backend='opencl')

        find_neighbor_cids = self._leaf_neighbor_operation(
            tree_src,
            args="uint2 *pbounds, int *cnt, int *neighbor_cids",
            setup="int offset=cnt[i];",
            operation="""
            if (pbounds[cid_src].s0 < pbounds[cid_src].s1)
                neighbor_cids[offset++] = cid_src;
            """,
            output_expr=""
        )
        find_neighbor_cids = profile_kernel(
            find_neighbor_cids, 'find_neighbor_cids', backend='opencl')
        find_neighbor_cids(tree_src.pbounds.dev,
                           neighbor_cid_count.dev, neighbor_cids.dev)
        return neighbor_cid_count, neighbor_cids

    # TODO?: 1D and 2D NNPS not properly supported here.
    # Just assuming the other spatial coordinates (y and z in case of 1D,
    # and z in case of 2D) are set to 0.
    def find_neighbor_lengths_elementwise(self, neighbor_cid_count,
                                          neighbor_cids, tree_src,
                                          neighbor_count):
        self.check_nnps_compatibility(tree_src)

        pa_gpu_dst = self.pa.gpu
        pa_gpu_src = tree_src.pa.gpu
        dtype = ctype_to_dtype(self.c_type)

        find_neighbor_counts = self.helper.get_kernel(
            'find_neighbor_counts_elementwise', sorted=self.sorted
        )
        find_neighbor_counts(self.unique_cids_map.dev, tree_src.pids.dev,
                             self.pids.dev,
                             self.cids.dev,
                             tree_src.pbounds.dev, self.pbounds.dev,
                             pa_gpu_src.x.dev, pa_gpu_src.y.dev,
                             pa_gpu_src.z.dev,
                             pa_gpu_src.h.dev,
                             pa_gpu_dst.x.dev, pa_gpu_dst.y.dev,
                             pa_gpu_dst.z.dev,
                             pa_gpu_dst.h.dev,
                             dtype(self.radius_scale),
                             neighbor_cid_count.dev,
                             neighbor_cids.dev,
                             neighbor_count.dev)

    def find_neighbors_elementwise(self, neighbor_cid_count, neighbor_cids,
                                   tree_src, start_indices, neighbors):
        self.check_nnps_compatibility(tree_src)

        pa_gpu_dst = self.pa.gpu
        pa_gpu_src = tree_src.pa.gpu

        dtype = ctype_to_dtype(self.c_type)

        find_neighbors = self.helper.get_kernel(
            'find_neighbors_elementwise', sorted=self.sorted)
        find_neighbors(self.unique_cids_map.dev, tree_src.pids.dev,
                       self.pids.dev,
                       self.cids.dev,
                       tree_src.pbounds.dev, self.pbounds.dev,
                       pa_gpu_src.x.dev, pa_gpu_src.y.dev, pa_gpu_src.z.dev,
                       pa_gpu_src.h.dev,
                       pa_gpu_dst.x.dev, pa_gpu_dst.y.dev, pa_gpu_dst.z.dev,
                       pa_gpu_dst.h.dev,
                       dtype(self.radius_scale),
                       neighbor_cid_count.dev,
                       neighbor_cids.dev,
                       start_indices.dev,
                       neighbors.dev)

    def _is_valid_nnps_wgs(self):
        # Max work group size can only be found by building the
        # kernel.
        try:
            find_neighbor_counts = self.helper.get_kernel(
                'find_neighbor_counts', sorted=self.sorted, wgs=self.leaf_size
            )

            find_neighbor = self.helper.get_kernel(
                'find_neighbors', sorted=self.sorted, wgs=self.leaf_size
            )
        except DeviceWGSException:
            return False
        else:
            return True

    def find_neighbor_lengths(self, neighbor_cid_count, neighbor_cids,
                              tree_src, neighbor_count,
                              use_partitions=False):
        self.check_nnps_compatibility(tree_src)

        wgs = self.leaf_size
        pa_gpu_dst = self.pa.gpu
        pa_gpu_src = tree_src.pa.gpu
        dtype = ctype_to_dtype(self.c_type)

        def find_neighbor_counts_for_partition(partition_cids, partition_size,
                                               partition_wgs, q=None):
            find_neighbor_counts = self.helper.get_kernel(
                'find_neighbor_counts', sorted=self.sorted, wgs=wgs
            )
            find_neighbor_counts(partition_cids.dev, tree_src.pids.dev,
                                 self.pids.dev,
                                 self.cids.dev,
                                 tree_src.pbounds.dev, self.pbounds.dev,
                                 pa_gpu_src.x.dev, pa_gpu_src.y.dev,
                                 pa_gpu_src.z.dev,
                                 pa_gpu_src.h.dev,
                                 pa_gpu_dst.x.dev, pa_gpu_dst.y.dev,
                                 pa_gpu_dst.z.dev,
                                 pa_gpu_dst.h.dev,
                                 dtype(self.radius_scale),
                                 neighbor_cid_count.dev,
                                 neighbor_cids.dev,
                                 neighbor_count.dev,
                                 gs=(partition_wgs * partition_size,),
                                 ls=(partition_wgs,),
                                 queue=(get_queue() if q is None else q))

        if use_partitions and wgs > 32:
            if wgs < 128:
                wgs1 = 32
            else:
                wgs1 = 64

            m1, n1 = self.get_leaf_size_partitions(0, wgs1)

            find_neighbor_counts_for_partition(m1, n1, min(wgs, wgs1))
            m2, n2 = self.get_leaf_size_partitions(wgs1, wgs)
            find_neighbor_counts_for_partition(m2, n2, wgs)
        else:
            find_neighbor_counts_for_partition(
                self.unique_cids, self.unique_cid_count, wgs)

    def find_neighbors(self, neighbor_cid_count, neighbor_cids, tree_src,
                       start_indices, neighbors, use_partitions=False):
        self.check_nnps_compatibility(tree_src)

        wgs = self.leaf_size if self.leaf_size % 32 == 0 else \
            self.leaf_size + 32 - self.leaf_size % 32
        pa_gpu_dst = self.pa.gpu
        pa_gpu_src = tree_src.pa.gpu
        dtype = ctype_to_dtype(self.c_type)

        def find_neighbors_for_partition(partition_cids, partition_size,
                                         partition_wgs, q=None):
            find_neighbors = self.helper.get_kernel('find_neighbors',
                                                    sorted=self.sorted,
                                                    wgs=wgs)
            find_neighbors(partition_cids.dev, tree_src.pids.dev,
                           self.pids.dev,
                           self.cids.dev,
                           tree_src.pbounds.dev, self.pbounds.dev,
                           pa_gpu_src.x.dev, pa_gpu_src.y.dev, pa_gpu_src.z.dev,
                           pa_gpu_src.h.dev,
                           pa_gpu_dst.x.dev, pa_gpu_dst.y.dev, pa_gpu_dst.z.dev,
                           pa_gpu_dst.h.dev,
                           dtype(self.radius_scale),
                           neighbor_cid_count.dev,
                           neighbor_cids.dev,
                           start_indices.dev,
                           neighbors.dev,
                           gs=(partition_wgs * partition_size,),
                           ls=(partition_wgs,),
                           queue=(get_queue() if q is None else q))

        if use_partitions and wgs > 32:
            if wgs < 128:
                wgs1 = 32
            else:
                wgs1 = 64

            m1, n1 = self.get_leaf_size_partitions(0, wgs1)
            fraction = (n1 / int(self.unique_cid_count))

            if fraction > 0.3:
                find_neighbors_for_partition(m1, n1, wgs1)
                m2, n2 = self.get_leaf_size_partitions(wgs1, wgs)
                assert (n1 + n2 == self.unique_cid_count)
                find_neighbors_for_partition(m2, n2, wgs)
                return
        else:
            find_neighbors_for_partition(
                self.unique_cids, self.unique_cid_count, wgs)

    def check_nnps_compatibility(self, tree):
        """Check if tree types and parameters are compatible for NNPS

        Two trees must satisfy a few conditions so that NNPS can be performed
        on one tree using the other as reference. In this case, the following
        conditions must be satisfied -

        1) Currently both should be instances of point_tree.PointTree
        2) Both must have the same sortedness
        3) Both must use the same floating-point datatype
        4) Both must have the same leaf sizes
        """
        if not isinstance(tree, PointTree):
            raise IncompatibleTreesException(
                "Both trees must be of the same type for NNPS"
            )

        if self.sorted != tree.sorted:
            raise IncompatibleTreesException(
                "Tree sortedness need to be the same for NNPS"
            )

        if self.c_type != tree.c_type or self.use_double != tree.use_double:
            raise IncompatibleTreesException(
                "Tree floating-point data types need to be the same for NNPS"
            )

        if self.leaf_size != tree.leaf_size:
            raise IncompatibleTreesException(
                "Tree leaf sizes need to be the same for NNPS (%d != %d)" % (
                    self.leaf_size, tree.leaf_size)
            )

        return



"""Shallow water sloshing

The case is as described in "Bouscasse, B., Antuono, M.,
Colagrossi, A., & Lugni, C. (2013). Numerical and Experimental
Investigation of Nonlinear Shallow Water Sloshing,
International Journal of Nonlinear Sciences and Numerical Simulation,
14(2), 123-138"

DOI:10.1515/ijnsns-2012-0100

This case is same as the Series 5 (Large amplitude sway motion with
a small water depth) case mentioned in the above paper.
"""

from math import cos, pi, exp, sqrt, tanh, sin

import os
import numpy as np

from pysph.base.utils import get_particle_array
from pysph.base.kernels import CubicSpline
from pysph.solver.solver import Solver
from pysph.solver.application import Application

from pysph.sph.integrator import EPECIntegrator

from pysph.sph.equation import Equation
from pysph.tools.geometry import get_2d_tank, get_2d_block
from pysph.sph.wc.edac import EDACScheme, EDACStep


L = 1  # Length of tank
h = 0.03 * L  # Height of water
amp = 2.333 * h  # Amplitude of oscillation

u_max = 1.3
c0 = 10.0 * u_max

dx = h/6
hdx = 1.2

h0 = hdx * dx
n_layers = 4

tf = 33
rho = 1000.0

k = pi/L
omega_r = sqrt(9.81*k*tanh(k*h))
omega = omega_r*1.231


class HorizontalExcitation(Equation):
    def __init__(self, dest, sources):
        L = 1
        h = 0.03 * L
        self.amp = 2.333 * h

        k = pi/L
        omega_r = sqrt(9.81*k*tanh(k*h))
        self.omega = omega_r*1.231
        super(HorizontalExcitation, self).__init__(dest, sources)

    def initialize(self, d_idx, d_au, t, d_rho):
        amp = self.amp
        omega = self.omega

        d_au[d_idx] += amp * (omega) * (omega) * sin(omega*t)


class SloshingTank(Application):
    def add_user_options(self, group):
        group.add_argument(
            '--dx', action='store', type=float, dest='dx', default=dx,
            help='Particle spacing.'
        )
        group.add_argument(
            '--hdx', action='store', type=float, dest='hdx', default=hdx,
            help='Specify the hdx factor where h = hdx * dx.'
        )

    def consume_user_options(self):
        self.hdx = self.options.hdx
        self.dx = self.options.dx
        self.h0 = self.hdx * self.dx

    def create_particles(self):
        dx = self.dx
        h0 = self.h0
        volume = dx * dx
        m = rho * volume

        xt, yt = get_2d_tank(
            dx=dx, length=L, height=0.2*L,
            num_layers=n_layers, base_center=[L/2, -dx]
            )

        xf, yf = get_2d_block(
            dx=dx, length=L - 2*dx, height=h, center=[L/2, h/2]
            )

        fluid = get_particle_array(name='fluid', x=xf, y=yf, h=h0,
                                   m=m, rho=rho)
        solid = get_particle_array(name='solid', x=xt, y=yt, h=h0,
                                   m=m, rho=rho)

        fluid.u = -amp * omega * np.ones_like(xf)

        self.scheme.setup_properties([fluid, solid])
        particles = [fluid, solid]
        return particles

    def create_solver(self):
        kernel = CubicSpline(dim=2)
        integrator = EPECIntegrator(fluid=EDACStep())

        dt = 0.125 * self.h0 / c0
        self.scheme.configure(h=self.h0)
        solver = Solver(kernel=kernel, dim=2, integrator=integrator,
                        tf=tf, dt=dt)

        return solver

    def create_scheme(self):
        s = EDACScheme(['fluid'], ['solid'], dim=2, rho0=rho, c0=c0, gy=-9.81,
                       alpha=0.0, nu=0.0, h=h0, clamp_p=True)
        return s

    def create_equations(self):
        eqns = self.scheme.get_equations()
        eqns[1].equations.insert(
            -1, HorizontalExcitation(dest='fluid', sources=None)
            )
        return eqns

    def post_process(self, info_fname):
        """
        Comparing wave height 0.05 m to the left of the tank
        """
        self.read_info(info_fname)

        if len(self.output_files) == 0:
            return

        from pysph.solver.utils import iter_output
        import matplotlib.pyplot as plt
        from cyarray.api import UIntArray
        from pysph.base import nnps

        files = self.output_files
        wave_height, t = [], []
        for sd, arrays1, arrays2 in iter_output(files, 'fluid', 'solid'):
            t.append(sd['t'])
            x_left = arrays2.x.min()
            probe_x = x_left + 0.05 + self.dx * (n_layers - 1)
            probe_y = np.linspace(0, arrays1.y.max(), 50)
            probe = get_particle_array(
                x=probe_x*np.ones_like(probe_y), y=probe_y
                )
            pa_list = [arrays1, probe]
            nps = nnps.LinkedListNNPS(dim=2, particles=pa_list, radius_scale=1)
            src_index, dst_index = 0, 1
            nps.set_context(src_index=0, dst_index=1)
            nbrs = UIntArray()
            wh = 0.0
            for i in range(len(probe_y)):
                nps.get_nearest_particles(src_index, dst_index, i, nbrs)
                neighbours = nbrs.get_npy_array()
                for j in range(len(neighbours)):
                    if arrays1.y[neighbours[j]] > wh:
                        wh = arrays1.y[neighbours[j]]

            wave_height.append(wh)

        wave_height = np.array(wave_height)
        wave_height = wave_height - h

        import pysph.examples.st_exp_data as st

        exp_t, exp_H = st.get_bouscasse_data()
        T = 2*pi/omega
        t = np.array(t)/T
        plt.plot(t, wave_height, label="Computed")
        plt.scatter(exp_t, exp_H, label="Experiment: Bouscasse et al, 2013")
        plt.xlabel("t/T")
        plt.ylabel("Wave Height (m)")
        plt.xlim(8.5, 11)
        plt.legend()
        plt.savefig(os.path.join(self.output_dir, 'H_vs_t.png'))
        plt.show()


if __name__ == '__main__':
    app = SloshingTank()
    app.run()
    app.post_process(app.info_filename)



"""Method of characteristic proposed by

    - Lastiwka, Martin, Mihai Basa, and Nathan J. Quinlan. "Permeable and
      non‐reflecting boundary conditions in SPH." International journal for
      numerical methods in fluids 61.7 (2009): 709-724.

"""

from pysph.sph.equation import Equation
from pysph.sph.integrator import PECIntegrator
from pysph.sph.bc.inlet_outlet_manager import InletOutletManager
from pysph.sph.wc.edac import EDACScheme
import numpy


class SimpleInletOutlet(InletOutletManager):
    def add_io_properties(self, pa, scheme=None):
        DEFAULT_PROPS = [
            'disp', 'ioid', 'Bp', 'A', 'wij', 'po', 'uho', 'vho', 'who',
            'Buh', 'Bvh', 'Bwh', 'x0', 'y0', 'z0', 'uhat', 'vhat', 'what',
            'xn', 'yn', 'zn', 'J1', 'J2u']
        STRIDE_DATA = {
            'A': 16, 'Bu': 4, 'Bv': 4, 'Bw': 4, 'Bp': 4, 'Brho': 4, 'uo':
            4, 'vo': 4, 'wo': 4, 'po': 4, 'rhoo': 4, 'Bau': 4, 'Bav': 4,
            'Baw': 4, 'auo': 4, 'avo': 4, 'awo': 4, 'Buh': 4, 'Bvh': 4,
            'Bwh': 4, 'uho': 4, 'vho': 4, 'who': 4}
        for prop in DEFAULT_PROPS:
            if prop in STRIDE_DATA:
                pa.add_property(prop, stride=STRIDE_DATA[prop])
            else:
                pa.add_property(prop)

        pa.add_constant('uref', 0.0)
        pa.add_constant('avgj2u', 0.0)
        pa.add_constant('avgj1', 0.0)

    def get_stepper(self, scheme, cls, edactvf=False):
        from pysph.sph.bc.inlet_outlet_manager import (
            InletStep, OutletStep)
        steppers = {}
        if (cls == PECIntegrator):
            if isinstance(scheme, EDACScheme):
                for inlet in self.inlets:
                    steppers[inlet] = InletStep()
                for outlet in self.outlets:
                    steppers[outlet] = OutletStep()
                for g_inlet in self.ghost_inlets:
                    steppers[g_inlet] = InletStep()
                self.active_stages = [2]

        return steppers

    def get_equations(self, scheme=None, summation_density=False,
                      edactvf=False):
        from pysph.sph.equation import Group
        from pysph.sph.bc.interpolate import (
            UpdateMomentMatrix, EvaluateUhat, EvaluateP, ExtrapolateUhat,
            ExtrapolateP, CopyUhatFromGhost, CopyPFromGhost)
        from pysph.sph.bc.inlet_outlet_manager import (
            UpdateNormalsAndDisplacements, CopyNormalsandDistances)

        equations = []
        g00 = []
        i = 0
        for info in self.inletinfo:
            g00.append(UpdateNormalsAndDisplacements(
                dest=info.pa_name, sources=None, xn=info.normal[0],
                yn=info.normal[1], zn=info.normal[2], xo=info.refpoint[0],
                yo=info.refpoint[1], zo=info.refpoint[2]
                ))
            g00.append(CopyNormalsandDistances(
                dest=self.inlet_pairs[info.pa_name], sources=[info.pa_name]))
            i = i + 1

        for fluid in self.fluids:
            g00.append(EvalauteCharacterisctics(
                dest=fluid, sources=None, c0=10.0, u0=1.0, v0=0.0,
                p0=1.0, rho0=1000.0))

        equations.append(Group(equations=g00, real=False))

        g02 = []
        for name in self.ghost_inlets:
            g02.append(UpdateMomentMatrix(
                dest=name, sources=self.fluids, dim=self.dim))

        equations.append(Group(equations=g02, real=False))

        g03 = []
        for name in self.ghost_inlets:
            g03.append(EvaluateUhat(dest=name, sources=self.fluids,
                                    dim=self.dim))
            g03.append(EvaluateP(dest=name, sources=self.fluids,
                                 dim=self.dim))
        for name in self.outlets:
            g03.append(EvalauteNumberdensity(dest=name, sources=self.fluids))
            g03.append(ShepardInterpolateCharacteristics(dest=name,
                                                         sources=self.fluids))

        equations.append(Group(equations=g03, real=False))

        g04 = []
        for name in self.ghost_inlets:
            g04.append(ExtrapolateUhat(dest=name, sources=None))
            g04.append(ExtrapolateP(dest=name, sources=None))
        for name in self.outlets:
            g04.append(EvaluatePropertyfromCharacteristics(
                dest=name, sources=None, c0=10.0, u0=1.0, v0=0.0,
                p0=1.0, rho0=1000.0))

        equations.append(Group(equations=g04, real=False))

        g05 = []
        for io in self.inlet_pairs.keys():
            g05.append(CopyUhatFromGhost(
                    dest=io, sources=[self.inlet_pairs[io]]))
            g05.append(CopyPFromGhost(
                dest=io, sources=[self.inlet_pairs[io]]))

        equations.append(Group(equations=g05, real=False))

        g06 = []
        for inlet in self.inletinfo:
            for eqn in inlet.equations:
                g06.append(eqn)
        for outlet in self.outletinfo:
            for eqn in outlet.equations:
                g06.append(eqn)

        equations.append(Group(equations=g06, real=False))

        return equations


class EvalauteCharacterisctics(Equation):
    def __init__(self, dest, sources, c0, rho0, u0, p0, v0):
        self.c0 = c0
        self.rho0 = rho0
        self.p0 = p0
        self.u0 = u0
        self.v0 = v0

        super(EvalauteCharacterisctics, self).__init__(dest, sources)

    def initialize(self, d_idx, d_u, d_v, d_p, d_rho, d_J1, d_J2u):
        a = self.c0
        rhoref = self.rho0
        uref = self.u0
        pref = self.p0

        d_J1[d_idx] = -a**2 * (d_rho[d_idx] - rhoref) + (d_p[d_idx] - pref)
        d_J2u[d_idx] = d_rho[d_idx] * a * (d_u[d_idx] - uref) + (d_p[d_idx] -
                                                                 pref)


class EvalauteNumberdensity(Equation):
    def initialize(self, d_idx, d_wij):
        d_wij[d_idx] = 0.0

    def loop(self, d_idx, d_wij, WIJ):
        d_wij[d_idx] += WIJ


class ShepardInterpolateCharacteristics(Equation):
    def initialize(self, d_idx, d_J1, d_J2u):
        d_J1[d_idx] = 0.0
        d_J2u[d_idx] = 0.0

    def loop(self, d_idx, d_J1, d_J2u, s_J1, s_J2u, WIJ, s_idx):
        d_J1[d_idx] += s_J1[s_idx] * WIJ
        d_J2u[d_idx] += s_J2u[s_idx] * WIJ

    def post_loop(self, d_idx, d_J1, d_J2u, d_wij, d_avgj2u, d_avgj1):
        if d_wij[d_idx] > 1e-14:
            d_J1[d_idx] /= d_wij[d_idx]
            d_J2u[d_idx] /= d_wij[d_idx]
        else:
            d_J2u[d_idx] = d_avgj2u[0]
            d_J1[d_idx] = d_avgj1[0]

    def reduce(self, dst, t, dt):
        dst.avgj2u[0] = numpy.average(dst.J2u[dst.wij > 0.0001])
        dst.avgj1[0] = numpy.average(dst.J1[dst.wij > 0.0001])


class EvaluatePropertyfromCharacteristics(Equation):
    def __init__(self, dest, sources, c0, rho0, u0, p0, v0):
        self.c0 = c0
        self.rho0 = rho0
        self.p0 = p0
        self.u0 = u0
        self.v0 = v0

        super(EvaluatePropertyfromCharacteristics, self).__init__(dest,
                                                                  sources)

    def initialize(self, d_idx, d_u, d_v, d_p, d_rho, d_J1, d_J2u):
        j1 = d_J1[d_idx]
        j2u = d_J2u[d_idx]

        d_rho[d_idx] = self.rho0 + (-j1 + 0.5 * j2u) / self.c0**2
        d_u[d_idx] = self.u0 + (j2u) / (2 * d_rho[d_idx] * self.c0)
        d_p[d_idx] = self.p0 + 0.5 * (j2u)



import numpy as np
import unittest
import pytest

import pysph.tools.geometry as G
from pysph.base.utils import get_particle_array


class TestGeometry(unittest.TestCase):

    def test_distance(self):
        point_1 = np.array([1.5, np.pi, np.exp(2.0)])
        point_2 = np.array([np.pi, np.exp(1), -1.9])
        d1 = G.distance(point_1, point_2)
        d2 = np.linalg.norm(point_1 - point_2)
        assert d1 == pytest.approx(d2)
        assert G.distance(point_1) == pytest.approx(np.linalg.norm(point_1))

    def test_distance_2d(self):
        point_1 = np.array([np.pi, np.exp(2.5)])
        point_2 = np.array([np.sqrt(2.4), np.pi])
        d1 = G.distance_2d(point_1, point_2)
        d2 = np.linalg.norm(point_1 - point_2)
        d3 = np.linalg.norm(point_1)
        assert d1 == pytest.approx(d2)
        assert G.distance_2d(point_1) == pytest.approx(d3)

    def test_matrix_exp(self):
        try:
            from scipy import linalg
            e = linalg.expm
            n = 3
            mat = np.array([[5.59, -6.5, 0.73],
                            [4.12, 6.4, -5.3],
                            [0.95, 3.54, -9.3]])
            assert np.allclose(e(mat), G.matrix_exp(mat))
        except ImportError:
            raise unittest.SkipTest('Scipy is not installed')

    def test_extrude(self):
        n = 15
        x = np.linspace(0.0, 2.5, n)
        y = np.linspace(-1.0, 4.0, n)
        dx = 0.05
        extrude_dist = 3.0
        x_new, y_new, z_new = G.extrude(x, y, dx, extrude_dist)
        assert np.allclose(x_new[:len(x)], x)
        assert np.allclose(x_new[-len(x):], x)
        assert np.allclose(y_new[:len(y)], y)
        assert np.allclose(y_new[-len(y):], y)

    def test_translate(self):
        n = 50
        x = np.linspace(1.0, 5.0, n)
        y = np.linspace(-5.0, 2.0, n)
        z = (x + y)**2 / 10.0
        x_shift = np.pi
        y_shift = np.exp(1)
        z_shift = np.sqrt(2.9)
        x_new, y_new, z_new = G.translate(x, y, z, x_shift, y_shift, z_shift)
        assert np.allclose(x_new - x_shift, x)
        assert np.allclose(y_new - y_shift, y)
        assert np.allclose(z_new - z_shift, z)

    def test_rotate(self):
        n = 15
        x = np.linspace(1.5, 10.0, n)
        y = np.linspace(5.0, 15.0, n)
        z = x * y / 10.0
        angle = 154.0
        theta = np.pi * angle / 180.0
        axis = np.array([1.0, np.sqrt(2.0), np.sqrt(3.0)])
        x_new, y_new, z_new = G.rotate(x, y, z, axis, angle)
        unit_vector = axis / np.linalg.norm(axis)
        mat = np.cross(np.eye(3), unit_vector * theta)
        rotation_matrix = G.matrix_exp(np.array(mat))
        test_points = []
        for xi, yi, zi in zip(x, y, z):
            point = np.array([xi, yi, zi])
            new = np.asarray(np.dot(rotation_matrix, point))
            test_points.append(new)
        test_points = np.asarray(test_points)
        x_test = test_points[:, 0]
        y_test = test_points[:, 1]
        z_test = test_points[:, 2]
        assert np.allclose(x_new, x_test)
        assert np.allclose(y_new, y_test)
        assert np.allclose(z_new, z_test)

    def test_get_2d_wall(self):
        dx = 0.05
        length = 3.0
        center = np.array([np.pi, 1.5])
        num_layers = 3
        up = True
        x, y = G.get_2d_wall(dx, center, length, num_layers, up)
        x_test = np.arange(-length / 2.0, length / 2.0, dx) + center[0]
        y_test = np.ones_like(x_test) * center[1]
        layer_length = int(len(x) / num_layers)
        value = 1 if up else -1
        assert np.allclose(x_test, x[:len(x_test)])
        assert np.allclose(y_test, y[:len(y_test)])
        assert np.allclose(x[:layer_length], x[-layer_length:])
        assert np.allclose(y[:layer_length],
                           y[-layer_length:] - value * (num_layers - 1) * dx)

    def test_get_2d_tank(self):
        dx = 0.05
        length = 5.0
        height = 4.0
        center = np.random.randn(2)
        num_layers = 4
        x, y = G.get_2d_tank(dx, center, length, height, num_layers, top=False,
                             staggered=False, outside=True)
        offset = (num_layers - 1)*dx
        xmin, xmax, ymin, ymax = min(x), max(x), min(y), max(y)
        lower_true = center[0] - 0.5*length, center[1]
        upper_true = center[0] + 0.5*length, center[1] + height
        assert (xmin + offset, ymin + offset) == pytest.approx((lower_true))
        assert (xmax - offset, ymax - offset) == pytest.approx((upper_true))

    def test_get_2d_circle(self):
        dx = 0.05
        radius = 3.0
        center = np.array([1.45, -0.43])
        x, y = G.get_2d_circle(dx, radius, center)
        xc, yc = x - center[0], y - center[1]
        assert not np.any(xc * xc + yc * yc) > radius * radius

    def test_get_2d_block(self):
        dx = 0.05
        length = 5.0
        height = 4.0
        center = np.array([0.4, 10.3])
        x, y = G.get_2d_block(dx, length, height, center)
        len_x = max(x) - min(x)
        len_y = max(y) - min(y)
        new_center = np.array(
            [(max(x) + min(x)) / 2.0, (max(y) + min(y)) / 2.0])
        assert len_x == pytest.approx(length)
        assert len_y == pytest.approx(height)
        assert np.allclose(new_center, center)

    def test_get_3d_sphere(self):
        dx = 0.23
        radius = 2.0
        center = np.array([np.pi, 1.5, np.exp(1.5)])
        x, y, z = G.get_3d_sphere(dx, radius, center)
        xs, ys, zs = x - center[0], y - center[1], z - center[2]
        assert not np.any(xs * xs + ys * ys + zs * zs) > radius * radius

    def test_get_3d_block(self):
        dx = 0.23
        length = 3.0
        height = 2.0
        depth = 1.5
        center = np.array([0.4, np.pi, -1.6])
        x, y, z = G.get_3d_block(dx, length, height, depth, center)
        len_x = max(x) - min(x)
        len_y = max(y) - min(y)
        len_z = max(z) - min(z)
        new_center = np.array([(max(x) + min(x)) / 2.0,
                               (max(y) + min(y)) / 2.0,
                               (max(z) + min(z)) / 2.0])
        assert len_x == pytest.approx(length)
        assert len_y == pytest.approx(height)
        assert len_z == pytest.approx(depth)
        assert np.allclose(center, new_center)

    def test_get_3d_hollow_cylinder(self):
        dx = 0.15
        radius = 2.0
        length = 3.0
        center = np.array([0.56, np.pi, np.sqrt(3.0)])
        num_layers = 5
        x, y, z = G.get_3d_hollow_cylinder(dx, radius, length, center,
                                           num_layers)
        xc, yc, zc = x - center[0], y - center[1], z - center[2]
        assert not np.any(xc * xc + yc * yc) > radius * radius
        assert not np.any(zc) > (length / 2.0)

    def test_get_4digit_naca_airfoil(self):
        dx = 0.1
        c = 1.5
        camber = '44'
        airfoils = ['0010', '2414', '4420', '3218', '2424']
        for airfoil in airfoils:
            x, y = G.get_4digit_naca_airfoil(dx, airfoil, c)
            count = 0
            t = 0.01 * float(airfoil[2:])
            for xi, yi in zip(x, y):
                yt = (0.2969 * np.sqrt(xi / c) - 0.1260 * (xi / c) -
                      0.3516 * ((xi / c)**2.) + 0.2843 * (xi / c)**3. -
                      0.1015 * ((xi / c)**4.))
                yt = yt * 5.0 * t
                if airfoil[:2] == '00':
                    if abs(yi) > yt + dx:
                        count += 1
                else:
                    m = 0.01 * float(airfoil[0])
                    p = 0.1 * float(airfoil[1])
                    if xi <= p * c:
                        yc = (m / (p**2.0)) * \
                            (2. * p * (xi / c) - (xi / c)**2.)
                        dydx = (2. * m / (p * p)) * (p - xi / c) / c
                    else:
                        yc = (m / ((1. - p) * (1. - p))) * \
                            (1. - 2. * p + 2. * p * (xi / c) - (xi / c)**2.)
                        dydx = (2. * m / ((1. - p) * (1. - p))) * \
                            (p - xi / c) / c
                    theta = np.arctan(dydx)
                    if yi >= 0.0:
                        yu = yc + yt * np.cos(theta)
                        if yi > yu + dx:
                            count += 1
                    else:
                        yl = yc - yt * np.cos(theta)
                        if yi < yl - dx:
                            count += 1
            assert count == 0

    def test_get_5digit_naca_airfoil(self):
        dx = 0.05
        c = 1.5
        series = ['210', '220', '230', '240', '250',
                  '221', '231', '241', '251']
        t = 12
        for s in series:
            airfoil = s + str(t)
            t = 0.01 * t
            x, y = G.get_5digit_naca_airfoil(dx, airfoil, c)
            count = 0
            m, k = G._get_m_k(s)
            for xi, yi in zip(x, y):
                yt = (0.2969 * np.sqrt(xi / c) - 0.1260 * (xi / c) -
                      0.3516 * ((xi / c)**2.) + 0.2843 * (xi / c)**3. -
                      0.1015 * ((xi / c)**4.))
                yt = 5.0 * t * yt
            xn = xi / c
            if xn <= m:
                yc = c * (k / 6.) * (xn**3. - 3. * m *
                                     xn * xn + m * m * (3. - m) * xn)
                dydx = (k / 6.) * (3. * xn * xn -
                                   6. * m * xn + m * m * (3. - m))
            else:
                yc = c * (k * (m**3.) / 6.) * (1. - xn)
                dydx = -(k * (m**3.) / 6.)
            theta = np.arctan(dydx)
            if yi >= 0.0:
                yu = yc + yt * np.cos(theta)
                if yi > yu + dx:
                    count += 1
            else:
                yl = yc - yt * np.cos(theta)
                if yi < yl - dx:
                    count += 1
            assert count == 0

    def test_get_naca_wing(self):
        dx = 0.15
        c = 1.5
        span = 3.0
        airfoils = ['23014', '22112', '0012', '2410']
        for airfoil in airfoils:
            x, y, z = G.get_naca_wing(dx, airfoil, span, c)
            assert not np.any(z) > (span / dx + dx)

    def test_remove_overlap_particles(self):
        dx_1 = 0.1
        dx_2 = 0.15
        length = 4.5
        height = 3.0
        radius = 2.0
        x1, y1 = G.get_2d_block(dx_1, length, height)
        x2, y2 = G.get_2d_circle(dx_2, radius)
        r1 = np.ones_like(x1) * 100.0
        m1 = r1 * dx_1 * dx_1
        h1 = np.ones_like(x1) * dx_1 * 1.5
        fluid = get_particle_array(name='fluid', x=x1, y=y1, h=h1, rho=r1,
                                   m=m1)
        r2 = np.ones_like(x2) * 100.0
        m2 = r2 * dx_2 * dx_2
        h2 = np.ones_like(x2) * dx_2 * 1.5
        solid = get_particle_array(name='solid', x=x2, y=y2, h=h2, rho=r2,
                                   m=m2)
        G.remove_overlap_particles(fluid, solid, dx_2, 2)
        x1 = fluid.x
        y1 = fluid.y
        count = 0
        for i in range(len(x1)):
            point = np.array([x1[i], y1[i]])
            dist = G.distance_2d(point)
            if dist <= radius:
                count += 1
        assert count == 0

    def test_remove_repeated_points(self):
        EPS = np.finfo(float).eps
        x = np.array([0, EPS, -1*EPS, 2*EPS, EPS/2, EPS*0.9, EPS*1.1, EPS, 10*EPS])
        y = np.zeros_like(x)
        z = np.zeros_like(x)
        xf, yf, zf = G.remove_repeated_points(x, y, z, EPS)
        np.testing.assert_array_equal(np.array([0, 10*EPS]), xf)
        np.testing.assert_array_equal(np.zeros(2), yf)
        np.testing.assert_array_equal(np.zeros(2), zf)

if __name__ == "__main__":
    unittest.main()



"""
Outlet boundary
"""

from pysph.sph.bc.inlet_outlet_manager import OutletBase
import numpy as np


class Outlet(OutletBase):
    def _get_ghost_xyz(self, x, y, z):
        xij = x - self.x
        yij = y - self.y
        zij = z - self.z

        disp = xij * self.xn + yij * self.yn + zij * self.zn
        x = x - 2 * disp * self.xn
        y = y - 2 * disp * self.yn
        z = z - 2 * disp * self.zn

        return x, y, z

    def update(self, time, dt, stage):
        if not self._init:
            self.initialize()
            self._init = True
        if stage in self.active_stages:
            props_to_copy = self.props_to_copy
            outlet_pa = self.outlet_pa
            source_pa = self.source_pa
            ghost_pa = self.ghost_pa

            self.io_eval = self._create_io_eval()
            self.io_eval.update()
            self.io_eval.evaluate()

            # adding particles to the destination array.
            io_id = source_pa.ioid
            cond = (io_id == 1)
            all_idx = np.where(cond)[0]
            pa_add = source_pa.extract_particles(
                all_idx, props=props_to_copy)
            outlet_pa.add_particles(**pa_add.get_property_arrays())

            if ghost_pa:
                if len(all_idx) > 0:
                    x, y, z = self._get_ghost_xyz(
                        pa_add.x, pa_add.y, pa_add.z)
                    pa_add.x = x
                    pa_add.y = y
                    pa_add.z = z
                    pa_add.u = -1. * pa_add.u
                    ghost_pa.add_particles(**pa_add.get_property_arrays())
            source_pa.remove_particles(all_idx)

            io_id = outlet_pa.ioid
            cond = (io_id == 2)
            all_idx = np.where(cond)[0]
            outlet_pa.remove_particles(all_idx)
            if ghost_pa:
                ghost_pa.remove_particles(all_idx)

            if self.callback is not None:
                self.callback(source_pa, outlet_pa)



import unittest

import numpy as np

from pysph.base import linalg3

class TestLinalg3(unittest.TestCase):

    def setUp(self):
        self.N = 10

    def assertMatricesEqual(self, result, expected, matrix, info, tol=1e-14):
        diff = result - expected
        msg = "Error for {info} matrix\n{matrix}\n"\
                "result:\n{result}\n"\
                "expected:\n{expected}\n"\
                "difference: {diff}".format(
            info=info, matrix=matrix,
            diff=diff, result=result, expected=expected
        )
        self.assertTrue(np.max(np.abs(diff)) < tol, msg)

    def _get_test_matrix(self):
        a = np.random.random((3,3))
        a += a.T
        return a

    def _get_difficult_matrix(self):
        a = np.identity(3, dtype=float)
        for p in range(3):
            p0 = int(p == 0)
            p2 = int(p == 2)
            a[p0][2-p2] = p0*1e-2
            a[2-p2][p0] = a[p0][2-p2]

        print(a)
        return a

    def _get_test_matrices(self):
        nasty = [
            [1.823886368900899e-169, -1.2724997010965309e-169, 0.0],
            [-1.2724997010965309e-169, -3.647772737801798e-169, 0.0],
            [0.0, 0.0, 0.0]
        ]

        data = [
            (np.zeros((3,3), dtype=float), 'zero'),
            (np.identity(3, dtype=float), 'identity'),
            (np.diag((1., 2., 3.)), 'diagonal'),
            (np.diag((2., 2., 1.)), 'diagonal repeated eigenvalues'),
            (np.ones((3,3), dtype=float), 'ones'),
            (self._get_test_matrix(), 'random'),
            (self._get_difficult_matrix(), 'difficult'),
            (np.asarray(nasty), 'nasty')

        ]
        return data

    def test_determinant(self):
        for i in range(self.N):
            self._check_determinant()

    def _check_determinant(self):
        # Given/When
        a = self._get_test_matrix()

        # Then
        self.assertAlmostEqual(
            linalg3.py_det(a), np.linalg.det(a), places=14
        )

    def test_eigen_values(self):
        # Given
        data = self._get_test_matrices()

        for matrix, info in data:
            # When
            result = linalg3.py_get_eigenvalues(matrix)
            result.sort()
            # Then
            expected = np.linalg.eigvals(matrix)
            expected.sort()
            self.assertMatricesEqual(result, expected, matrix, info)

    def test_eigen_decompose_eispack(self):
        # Given
        data = self._get_test_matrices()

        for matrix, info in data:
            # When
            val, vec = linalg3.py_eigen_decompose_eispack(matrix)

            # Then
            check = np.dot(np.asarray(vec), np.dot(np.diag(val), np.asarray(vec).T))
            self.assertMatricesEqual(check, matrix, vec, info)

    def test_get_eigenvalvec(self):
        # Given
        data = self._get_test_matrices()

        for matrix, info in data:
            # When
            val, vec = linalg3.py_get_eigenvalvec(matrix)

            # Then
            check = np.dot(np.asarray(vec), np.dot(np.diag(val), np.asarray(vec).T))
            self.assertMatricesEqual(check, matrix, vec, info)

    ##################################################################################
    # Transformation related tests.

    def test_transform_function(self):
        for i in range(self.N):
            self._check_transform_function()

    def _check_transform_function(self):
        # Given
        a = np.random.random((3,3))
        p = np.random.random((3,3))

        # When
        res = linalg3.py_transform(a, p)

        # Then
        expected = np.dot(p.T, np.dot(a, p))
        self.assertMatricesEqual(res, expected, (a,p), 'transform')

    def test_transform_diag_function(self):
        for i in range(self.N):
            self._check_transform_diag_function()

    def _check_transform_diag_function(self):
        # Given
        a = np.random.random(3)
        p = np.random.random((3,3))

        # When
        res = linalg3.py_transform_diag(a, p)

        # Then
        expected = np.dot(p.T, np.dot(np.diag(a), p))
        self.assertMatricesEqual(res, expected, (a,p), 'transform')

    def test_transform_diag_inv_function(self):
        for i in range(self.N):
            self._check_transform_diag_inv_function()

    def _check_transform_diag_inv_function(self):
        # Given
        a = np.random.random(3)
        p = np.random.random((3,3))

        # When
        res = linalg3.py_transform_diag_inv(a, p)

        # Then
        expected = np.dot(p, np.dot(np.diag(a), p.T))
        self.assertMatricesEqual(res, expected, (a,p), 'transform')

if __name__ == '__main__':
    unittest.main()
